#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
    This script generates code for in the typeables.py and action_key.py
    files.

"""


(vkey, lookup) = range(2)

key_map = (
    ("Lowercase letter keys", [
        (lookup, "a",                              "a alpha"),
        (lookup, "b",                              "b bravo"),
        (lookup, "c",                              "c charlie"),
        (lookup, "d",                              "d delta"),
        (lookup, "e",                              "e echo"),
        (lookup, "f",                              "f foxtrot"),
        (lookup, "g",                              "g golf"),
        (lookup, "h",                              "h hotel"),
        (lookup, "i",                              "i india"),
        (lookup, "j",                              "j juliet"),
        (lookup, "k",                              "k kilo"),
        (lookup, "l",                              "l lima"),
        (lookup, "m",                              "m mike"),
        (lookup, "n",                              "n november"),
        (lookup, "o",                              "o oscar"),
        (lookup, "p",                              "p papa"),
        (lookup, "q",                              "q quebec"),
        (lookup, "r",                              "r romeo"),
        (lookup, "s",                              "s sierra"),
        (lookup, "t",                              "t tango"),
        (lookup, "u",                              "u uniform"),
        (lookup, "v",                              "v victor"),
        (lookup, "w",                              "w whisky"),
        (lookup, "x",                              "x xray"),
        (lookup, "y",                              "y yankee"),
        (lookup, "z",                              "z zulu")]),
    ("Uppercase letter keys", [
        (lookup, "A",                              "A Alpha"),
        (lookup, "B",                              "B Bravo"),
        (lookup, "C",                              "C Charlie"),
        (lookup, "D",                              "D Delta"),
        (lookup, "E",                              "E Echo"),
        (lookup, "F",                              "F Foxtrot"),
        (lookup, "G",                              "G Golf"),
        (lookup, "H",                              "H Hotel"),
        (lookup, "I",                              "I India"),
        (lookup, "J",                              "J Juliet"),
        (lookup, "K",                              "K Kilo"),
        (lookup, "L",                              "L Lima"),
        (lookup, "M",                              "M Mike"),
        (lookup, "N",                              "N November"),
        (lookup, "O",                              "O Oscar"),
        (lookup, "P",                              "P Papa"),
        (lookup, "Q",                              "Q Quebec"),
        (lookup, "R",                              "R Romeo"),
        (lookup, "S",                              "S Sierra"),
        (lookup, "T",                              "T Tango"),
        (lookup, "U",                              "U Uniform"),
        (lookup, "V",                              "V Victor"),
        (lookup, "W",                              "W Whisky"),
        (lookup, "X",                              "X Xray"),
        (lookup, "Y",                              "Y Yankee"),
        (lookup, "Z",                              "Z Zulu")]),
    ("Number keys", [
        (lookup, "0",                              "0 zero"),
        (lookup, "1",                              "1 one"),
        (lookup, "2",                              "2 two"),
        (lookup, "3",                              "3 three"),
        (lookup, "4",                              "4 four"),
        (lookup, "5",                              "5 five"),
        (lookup, "6",                              "6 six"),
        (lookup, "7",                              "7 seven"),
        (lookup, "8",                              "8 eight"),
        (lookup, "9",                              "9 nine")]),
    ("Symbol keys", [
        (lookup, "!",                              "bang exclamation"),
        (lookup, "@",                              "at"),
        (lookup, "#",                              "hash"),
        (lookup, "$",                              "dollar"),
        (lookup, "%",                              "percent"),
        (lookup, "^",                              "caret"),
        (lookup, "&",                              "and ampersand"),
        (lookup, "*",                              "star asterisk"),
        (lookup, "(",                              "leftparen lparen"),
        (lookup, ")",                              "rightparen rparen"),
        (lookup, "-",                              "minus hyphen"),
        (lookup, "_",                              "underscore"),
        (lookup, "+",                              "plus"),
        (lookup, "`",                              "backtick"),
        (lookup, "~",                              "tilde"),
        (lookup, "[",                              "leftbracket lbracket"),
        (lookup, "]",                              "rightbracket rbracket"),
        (lookup, "{",                              "leftbrace lbrace"),
        (lookup, "}",                              "rightbrace rbrace"),
        (lookup, "\\",                             "backslash"),
        (lookup, "|",                              "bar"),
        (lookup, ":",                              "colon"),
        (lookup, ";",                              "semicolon"),
        (lookup, "'",                              "apostrophe singlequote squote"),
        (lookup, '"',                              "quote doublequote dquote"),
        (lookup, ",",                              "comma"),
        (lookup, ".",                              "dot"),
        (lookup, "/",                              "slash"),
        (lookup, "<",                              "lessthan leftangle langle"),
        (lookup, ">",                              "greaterthan rightangle rangle"),
        (lookup, "?",                              "question"),
        (lookup, "=",                              "equal equals")]),
    ("Whitespace and editing keys", [
        (vkey,   "win32con.VK_RETURN",             "enter"),
        (vkey,   "win32con.VK_TAB",                "tab"),
        (vkey,   "win32con.VK_SPACE",              "space"),
        (vkey,   "win32con.VK_BACK",               "backspace"),
        (vkey,   "win32con.VK_DELETE",             "delete del")]),
    ("Modifier keys", [
        (vkey,   "win32con.VK_SHIFT",              "shift"),
        (vkey,   "win32con.VK_CONTROL",            "control ctrl"),
        (vkey,   "win32con.VK_MENU",               "alt")]),
    ("Special keys", [
        (vkey,   "win32con.VK_ESCAPE",             "escape"),
        (vkey,   "win32con.VK_INSERT",             "insert"),
        (vkey,   "win32con.VK_PAUSE",              "pause"),
        (vkey,   "win32con.VK_LWIN",               "win"),
        (vkey,   "win32con.VK_APPS",               "apps popup")]),
    ("Navigation keys", [
        (vkey,   "win32con.VK_UP",                 "up"),
        (vkey,   "win32con.VK_DOWN",               "down"),
        (vkey,   "win32con.VK_LEFT",               "left"),
        (vkey,   "win32con.VK_RIGHT",              "right"),
        (vkey,   "win32con.VK_PRIOR",              "pageup pgup"),
        (vkey,   "win32con.VK_NEXT",               "pagedown pgdown"),
        (vkey,   "win32con.VK_HOME",               "home"),
        (vkey,   "win32con.VK_END",                "end")]),
    ("Number pad keys", [
        (vkey,   "win32con.VK_MULTIPLY",           "npmul"),
        (vkey,   "win32con.VK_ADD",                "npadd"),
        (vkey,   "win32con.VK_SEPARATOR",          "npsep"),
        (vkey,   "win32con.VK_SUBTRACT",           "npsub"),
        (vkey,   "win32con.VK_DECIMAL",            "npdec"),
        (vkey,   "win32con.VK_DIVIDE",             "npdiv"),
        (vkey,   "win32con.VK_NUMPAD0",            "numpad0 np0"),
        (vkey,   "win32con.VK_NUMPAD1",            "numpad1 np1"),
        (vkey,   "win32con.VK_NUMPAD2",            "numpad2 np2"),
        (vkey,   "win32con.VK_NUMPAD3",            "numpad3 np3"),
        (vkey,   "win32con.VK_NUMPAD4",            "numpad4 np4"),
        (vkey,   "win32con.VK_NUMPAD5",            "numpad5 np5"),
        (vkey,   "win32con.VK_NUMPAD6",            "numpad6 np6"),
        (vkey,   "win32con.VK_NUMPAD7",            "numpad7 np7"),
        (vkey,   "win32con.VK_NUMPAD8",            "numpad8 np8"),
        (vkey,   "win32con.VK_NUMPAD9",            "numpad9 np9")]),
    ("Function keys", [
        (vkey,   "win32con.VK_F1",                 "f1"),
        (vkey,   "win32con.VK_F2",                 "f2"),
        (vkey,   "win32con.VK_F3",                 "f3"),
        (vkey,   "win32con.VK_F4",                 "f4"),
        (vkey,   "win32con.VK_F5",                 "f5"),
        (vkey,   "win32con.VK_F6",                 "f6"),
        (vkey,   "win32con.VK_F7",                 "f7"),
        (vkey,   "win32con.VK_F8",                 "f8"),
        (vkey,   "win32con.VK_F9",                 "f9"),
        (vkey,   "win32con.VK_F10",                "f10"),
        (vkey,   "win32con.VK_F11",                "f11"),
        (vkey,   "win32con.VK_F12",                "f12"),
        (vkey,   "win32con.VK_F13",                "f13"),
        (vkey,   "win32con.VK_F14",                "f14"),
        (vkey,   "win32con.VK_F15",                "f15"),
        (vkey,   "win32con.VK_F16",                "f16"),
        (vkey,   "win32con.VK_F17",                "f17"),
        (vkey,   "win32con.VK_F18",                "f18"),
        (vkey,   "win32con.VK_F19",                "f19"),
        (vkey,   "win32con.VK_F20",                "f20"),
        (vkey,   "win32con.VK_F21",                "f21"),
        (vkey,   "win32con.VK_F22",                "f22"),
        (vkey,   "win32con.VK_F23",                "f23"),
        (vkey,   "win32con.VK_F24",                "f24")]),
    ("Multimedia keys", [
        (vkey,   "win32con.VK_VOLUME_UP",          "volumeup volup"),
        (vkey,   "win32con.VK_VOLUME_DOWN",        "volumedown voldown"),
        (vkey,   "win32con.VK_VOLUME_MUTE",        "volumemute volmute"),
        (vkey,   "win32con.VK_MEDIA_NEXT_TRACK",   "tracknext"),
        (vkey,   "win32con.VK_MEDIA_PREV_TRACK",   "trackprev"),
        (vkey,   "win32con.VK_MEDIA_PLAY_PAUSE",   "playpause"),
        (vkey,   "win32con.VK_BROWSER_BACK",       "browserback"),
        (vkey,   "win32con.VK_BROWSER_FORWARD",    "browserforward")]),
    )

print("---- Code for typeables.py")
for group_name, group_map in key_map:
    print("    # %s" % (group_name,))
    for key_type, key_value, key_names in group_map:
        for key_name in key_names.split():
            if key_type == lookup:
                value_code = "keyboard.get_typeable(char=%r)" % key_value
            elif key_type == vkey:
                value_code = ("Typeable(code=%s, name=%r)"
                              % (key_value, key_name))
            else:
                raise Exception("Invalid key type: {0!r} (for {1!r} {2!r})"
                                .format(key_type, key_value, key_names))
            print(('    "%s": %s%s,'
                   % (key_name, " " * (16-len(key_name)), value_code)))
    print("")

print("---- Code for documentation in action_key.py")
for group_name, group_map in key_map:
    parts = [group_name + ":"]
    for key_type, key_value, key_names in group_map:
        key_name_parts = []
        for key_name in key_names.split():
            key_name_parts.append("``" + key_name + "``")
        parts.append(" or ".join(key_name_parts) + ",")
    parts[-1] = parts[-1][:-1]  # Remove trailing comma.
    line = " -"
    for part in parts:
        concatenation = line + " " + part
        if len(concatenation) <= 72:
            line = concatenation
        else:
            print(line)
            line = "   " + part
    print(line)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
PlaySound action
============================================================================

"""

import winsound
from .action_base         import ActionBase, ActionError


#---------------------------------------------------------------------------

class PlaySound(ActionBase):

    def __init__(self, name=None, file=None):
        ActionBase.__init__(self)
        if name is not None:
            self._name = name
            self._flags = winsound.SND_ASYNC | winsound.SND_ALIAS
        elif file is not None:
            self._name = file
            self._flags = winsound.SND_ASYNC | winsound.SND_FILENAME

        self._str = str(self._name)

    def _execute(self, data=None):
        winsound.PlaySound(self._name, self._flags)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Loader
============================================================================

"""

import os.path
import glob
import time
import configparser
import collections
from collections import defaultdict

import dragonfly
from ...error                    import DragonflyError
from ...grammar.rule_base        import Rule
from ...grammar.context          import AppContext
from ...actions                  import ActionBase
from ...parser                   import Parser
from .dirmon                     import DirectoryMonitor
from .family                     import CommandFamily
from .state                      import StateBase as State
from .state                      import Transition
from .command                    import CompoundCommand
from .loader_parser              import CallElement


#===========================================================================
# Exception classes.

class LoaderError(DragonflyError):
    pass

class SyntaxError(LoaderError):
    pass


#===========================================================================
# Container base class similar to C's struct.

class ContainerBase(object):

    _attributes = ()

    def __init__(self, **kwargs):
        for name, default in self._attributes:
            if name in kwargs:       value = kwargs.pop(name)
            elif isinstance(default, collections.Callable):  value = default()
            else:                    value = default
            setattr(self, name, value)
        if kwargs:
            names = sorted(kwargs.keys())
            raise ValueError("Unknown keyword args: %s" % names)


#===========================================================================
# Input information storage classes.

class InfoPhase1(ContainerBase):

    class InfoPhase1Section(ContainerBase):
        _attributes = (
                       ("tag",           None),
                       ("items",         dict),
                      )

    _attributes = (
                   ("sections", lambda: defaultdict(InfoPhase1.InfoPhase1Section)),
                  )


class InfoPhase2(ContainerBase):

    class InfoPhase2Family(ContainerBase):
        _attributes = (
                       ("tag",           None),
                       ("name",          None),
                       ("description",   None),
                       ("states",        dict),
                       ("extras",        list),
                      )
    class InfoPhase2Choice(ContainerBase):
        _attributes = (
                       ("pairs",         None),
                      )

    _attributes = (
                   ("families", lambda: defaultdict(InfoPhase2.InfoPhase2Family)),
                   ("choices",  lambda: defaultdict(InfoPhase2.InfoPhase2Choice)),
                  )


#===========================================================================
# Config parser class.

class CommandConfigParser(configparser.RawConfigParser):
    """ Customized ConfigParser class for parsing command files. """

    def optionxform(self, option):
        """
            Return option-string unchanged.

            This overrides the default behavior of converting the 
            option-string to lowercase.

        """
        return option


#===========================================================================
# Structured container classes for storing config file input.

class ConfigSection(object):

    def __init__(self, name, items):
        self._section_name = name
        self._section_items = items
        self._parse_name(self._section_name)
        self._parse_items(self._section_items)

    def _parse_name(self, name):
        pass

    def _parse_items(self, items):
        pass

    def _split_name_type_tag(self, name):
        sep_index = name.find(":")
        if sep_index == -1:
            raise Exception("Invalid section name: %r" % name)
        section_type = name[:sep_index].strip().lower()
        section_tag = name[sep_index+1:].strip().lower()
        if not section_type or not section_tag:
            raise Exception("Invalid section name: %r" % name)
        return section_type, section_tag

    def _split_name_type_family_tag(self, name):
        section_type, family_section_tag = self._split_name_type_tag(name)
        sep_index = family_section_tag.find(".")
        if sep_index == -1:
            raise Exception("Invalid section name: %r" % name)
        family_tag = family_section_tag[:sep_index].strip().lower()
        if not family_tag:
            raise Exception("Invalid section name: %r" % name)
        return section_type, family_tag, family_section_tag

    def _build_items_dict(self, items, error_on_duplicate_key=True):
        items_dict = {}
        for key, value in items:
            if key in items_dict:
                if error_on_duplicate_key:
                    raise Exception("Duplicate key: %r" % key)
                continue
            items_dict[key] = value
        return items_dict

    def _unescape_spoken_form(self, escaped):
        escaped = escaped.strip()
        if escaped[0] != '"' or escaped[-1] != '"':
            raise Exception("Invalid spoken form: %r" % escaped)
        unescaped = escaped[1:-1]
        return unescaped


class FamilySection(ConfigSection):

    def _parse_name(self, name):
        section_type, family_tag = self._split_name_type_tag(name)
        self.tag = family_tag

    def _parse_items(self, items):
        items_dict = self._build_items_dict(items)

        # Parse "name" item, if present.
        if "name" in items_dict:
            name_string = items_dict.pop("name")
            self.name = name_string.strip()
        else:
            raise Exception("Family section %r missing name item."
                            % self._section_name)

        # Parse "description" item, if present.
        if "description" in items_dict:
            description_string = items_dict.pop("description")
            self.description = description_string.strip()
        else:
            self.description = None

        # Parse "length" item, if present.
        if "length" in items_dict:
            length_string = items_dict.pop("length")
            self.length = int(length_string)
        else:
            self.length = 4

        if items_dict:
            raise Exception("Family section contains invalid items:"
                            " %s" % (sorted(items_dict.keys()),))


class StateSection(ConfigSection):

    def _parse_name(self, name):
        section_type, family_tag, tag = self._split_name_type_family_tag(name)
        self.family_tag = family_tag
        self.tag = tag

    def _parse_items(self, items):
        items_dict = self._build_items_dict(items)

        # Parse "name" item, if present.
        if "name" in items_dict:
            self.name = items_dict.pop("name")
        else:
            self.name = self.tag

        # Parse "include" item, if present.
        if "include" in items_dict:
            include_string = items_dict.pop("include")
            include_names = [s.strip() for s in include_string.split(",")]
            self.include = include_names
        else:
            self.include = ()

        # Parse "context" item, if present.
        if "context" in items_dict:
            context_string = items_dict.pop("context")
            self.context = context_string.strip()
        else:
            raise Exception("State section %r missing context item."
                            % self._section_name)

        # Parse "next" item, if present.
        if "next" in items_dict:
            next_string = items_dict.pop("next")
            self.next = next_string.strip()
        else:
            self.next = None

        self.commands = []
        for spoken, spec in items_dict.items():
            spoken = self._unescape_spoken_form(spoken)
            self.commands.append((spoken, spec))


class ExtrasSection(ConfigSection):

    def _parse_name(self, name):
        section_type, family_tag = self._split_name_type_tag(name)
        self.tag = family_tag

    def _parse_items(self, items):
        self.extras = []
        for item in items:
            key, spec = item
            self.extras.append((key, spec))


class ChoiceSection(ConfigSection):

    def _parse_name(self, name):
        section_type, family_tag = self._split_name_type_tag(name)
        self.tag = family_tag

    def _parse_items(self, items):
        self.choices = []
        for item in items:
            spoken, spec = item
            self.choices.append((spoken, spec))


class InputSpec(object):
        pass


class InputInfo(ContainerBase):

    _attributes = (
                   ("family_sections",  None),
                   ("family_states",    None),
                   ("family_extras",    None),
                   ("choice_sections",  None),
                  )


#===========================================================================

class Loader(object):

    _extras_factories = {}

    @classmethod
    def register_extras_factory(cls, extra_type, factory):
        cls._extras_factories[extra_type] = factory

    def __init__(self):
        self._dirmon = DirectoryMonitor()
        self._loaded = False
        self._families = []

    def add_directories(self, *directories):
        """
            Add one or more directories to be monitored.

            Parameter:
              - directories (sequence of *str* or *unicode*) --
                the new directories to be monitored.

        """
        self._dirmon.add_directories(*directories)

    #-----------------------------------------------------------------------

    def load(self):
        """ Load families found in monitored directories. """
        self._loaded = True

        # Find files to be loaded.
        directories = self._dirmon.directories
        command_files = self._find_command_files(directories)
        library_files = self._find_library_files(directories)
#        print "command files", command_files
#        print "library files", library_files

        # Read and parse command files.
        input_spec = self._parse_command_files(command_files)

        # Load family objects into engine.
        for family in self._families:
            family.load()

    def unload(self):
        """ Unload all commands. """
        for family in self._families:
            family.unload()
        self._loaded = False

    def check_for_changes(self):
        """
            Check for changes and reload if necessary.

            Return *True* if files were reloaded; otherwise *False*.

        """
        if not self._loaded or not self._dirmon.is_modified():
            return False
        self.unload()
        self.load()
        return True

    #-----------------------------------------------------------------------

    def _find_command_files(self, directories):
        return self._glob_files_in_directories(directories, "*.txt")

    def _find_library_files(self, directories):
        return self._glob_files_in_directories(directories, "*.py")

    def _glob_files_in_directories(self, directories, pattern):
        files = []
        for directory in directories:
            directory_pattern = os.path.join(directory, pattern)
            files.extend(glob.glob(directory_pattern))
        return files

        for directory in directories:
            for filename in os.listdir(directory):
                extension = os.path.splitext(filename)[1]
                if extension.lower() == ".txt":
                    files.append(os.path.join(directory, filename))

    def _parse_command_files(self, command_files):
        # Create config parser and read command files.
        config = CommandConfigParser()
        for path in command_files:
            config.read(path)

        dispatchers = {
                       "family:":   FamilySection,
                       "state:":    StateSection,
                       "extras:":   ExtrasSection,
                       "choice:":   ChoiceSection,
                      }

        # Iterate over all input sections and process each according
        #  to the section name's prefix.
        sections = []
        for section_name in config.sections():
            items = config.items(section_name)
            cooked_name = section_name.strip().lower()
            section_instance = None
            for prefix, section_class in dispatchers.items():
                if cooked_name.startswith(prefix):
                    section_instance = section_class(section_name, items)
                    break
            if not section_instance:
                raise Exception("Invalid section: %r" % section_name)
            sections.append(section_instance)

        # Iterate over all processed section objects and handle each
        #  according to its type.
        family_sections = {}
        family_states   = {}
        family_extras   = {}
        choice_sections = {}
        for section in sections:
            if isinstance(section, FamilySection):
                family_sections[section.tag] = section
            elif isinstance(section, StateSection):
                states = family_states.setdefault(section.family_tag, {})
                states[section.tag] = section
            elif isinstance(section, ExtrasSection):
                family_extras[section.tag] = section
            elif isinstance(section, ChoiceSection):
                choice_sections[section.tag] = section
            else:
                raise Exception("Invalid section type: %r" % section_name)
        input_info = InputInfo(
                               family_sections  = family_sections,
                               family_states    = family_states,
                               family_extras    = family_extras,
                               choice_sections  = choice_sections,
                              )

        # Iterate over all family sections and construct each family.
        families = []
        for family_section in family_sections.values():
            print("constructing family", family_section.tag)
            family = CommandFamily(name=family_section.name)

            extras_section = family_extras[family_section.tag]
            print("  constructing extras", extras_section.tag, extras_section._section_name)
            self._build_family_extras(family, extras_section, input_info)

            states_by_tag = self._init_family_states(family, family_section, input_info)
            for state_section in family_states[family_section.tag].values():
                print("  constructing state", state_section.tag)
#                self._build_family_state(family, state_section, states_by_tag, input_info)

            families.append(family)
        return families

    def _build_family_extras(self, family, extras_section, input_info):
        element = CallElement()
        parser = Parser(element)
        for key, spec in extras_section.extras:
            print("building extra", key, spec)

            # Parse the input spec.
            output = parser.parse(spec)
            output.name = key
            print("output:", output)
            if not output:
                raise SyntaxError("Invalid extra %r: %r" % (key, spec))

            # Look for an appropriate extras factory and let it
            #  build the extra element.
            if output.function not in self._extras_factories:
                raise SyntaxError("Unknown extra type %r in %r" % (output.function, spec))
            factory = self._extras_factories[output.function]
            extra = factory.build(output)

            family.add_extras(extra)

    def _init_family_states(self, family, family_section, input_info):
        sections = list(input_info.family_states[family_section.tag].values())
        states = []
        states_by_tag = {}
        for section in sections:
            state = State(section.name)
            states.append(state)
            states_by_tag[section.tag] = state
        family.add_states(*states)
        return states_by_tag

    def _build_family_state_phase2(self, family, state_section, states_by_tag, input_info):
        state = family.states[state_section.tag]

        context = self._build_state_context(state_section.context)
#        state.set_context(context)

        for state_name in state_section.include:
            included_state = states_by_tag[state_name]
            state.include_states(included_state)

        extras = []

        for spoken, spec in state_section.commands:
            command, poststate_name = self._build_state_command(spoken, spec, extras)
            if not poststate_name:
                poststate = state
            else:
                poststate = states_by_tag[poststate_name]
            transition = Transition(command, state, poststate)
            state.add_transitions(transition)

    def _build_state_context(self, context_spec):
        pass

    def _build_state_command(self, spoken_form, command_spec, extras):
        # Prepare namespace in which to evaluate command_spec.
        namespace = {
#                     "__file__":  path,
#                     "library":   library,
                    }
        for (key, value) in dragonfly.__dict__.items():
            if isinstance(value, type) and issubclass(value, ActionBase):
                namespace[key] = value
        namespace["Repeat"] = dragonfly.Repeat

        # Evaluate command specification.
        expression = "(%s)" % command_spec
        action = eval(expression, namespace, namespace)

        # Wrapup action in extras in a compound command.
        result = (CompoundCommand(spoken_form, action, extras=extras), None)
        return result


#===========================================================================

class ExtrasFactoryBase(object):

    def __init__(self):
        pass

    def build(self, call_info):
        pass


#---------------------------------------------------------------------------

class DictationExtrasFactory(ExtrasFactoryBase):

    def build(self, call_info):
        name = call_info.name
        if call_info.arguments:
            raise SyntaxError("Invalid arguments for dictation extra: %r"
                              % (call_info.arguments,))
        print("just build", dragonfly.Dictation(name))
        return dragonfly.Dictation(name)

Loader.register_extras_factory("dictation", DictationExtrasFactory())


#---------------------------------------------------------------------------

class IntegerExtrasFactory(ExtrasFactoryBase):

    def build(self, call_info):
        name = call_info.name
        min = call_info.arguments.pop().value
        max = call_info.arguments.pop().value
        if call_info.arguments:
            raise SyntaxError("Invalid arguments for integer extra: %r"
                              % (call_info.arguments,))
        print("just build", dragonfly.Integer(name=name, min=min, max=max))
        return dragonfly.Integer(name=name, min=min, max=max)

Loader.register_extras_factory("integer", IntegerExtrasFactory())


#===========================================================================

class ExtrasFactoryError(Exception):
    pass

class ContextFactoryError(Exception):
    pass

class ActionFactoryError(Exception):
    pass

class ContextFactoryBase(object):
    pass

class ActionFactoryBase(object):
    pass



# -*- encoding: windows-1252 -*-
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#


import unittest
from dragonfly.actions.action_text import Text
from dragonfly.actions.action_paste import Paste
from dragonfly.actions.action_mimic import Mimic


#===========================================================================

# This test fails because the keyboard interface currently
# cannot type certain characters, such as accented vowels.
#
#class TestNonAsciiText(unittest.TestCase):
#
#    def test_non_ascii_text(self):
#        """ Test handling of non-ASCII characters in Text action. """
#
#        action = Text("touché")
#        self.assertEqual(str(action), "Text(%r)" % ("touché",))


class TestNonAsciiPaste(unittest.TestCase):

    def test_non_ascii_paste(self):
        """ Test handling of non-ASCII characters in Paste action. """

        action = Paste("touché")
        self.assertEqual(str(action), "Paste(%r)" % ("touché",))


class TestNonAsciiMimic(unittest.TestCase):

    def test_non_ascii_mimic(self):
        """ Test handling of non-ASCII characters in Mimic action. """

        action = Mimic("touché")
        self.assertEqual(str(action), "Mimic(%r)" % ("touché",))


#===========================================================================

if __name__ == "__main__":
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements a Win32 control base class.
"""


#---------------------------------------------------------------------------

import sys
import ctypes
import struct
import winxpgui as win32gui
import win32api
import win32con
import os


#---------------------------------------------------------------------------

class ControlBase(object):


    #-----------------------------------------------------------------------

    _next_id = 1024

    @classmethod
    def _get_next_id(cls):
        ControlBase._next_id += 1
        return ControlBase._next_id - 1

    _message_names = {
                      "on_command":      win32con.WM_COMMAND,
                     }


    #-----------------------------------------------------------------------

    def __init__(self, parent, flavor, text, size, style, **kwargs):
        self._parent = parent
        self._flavor = flavor
        self._text = text
        self._size = size
        self._style = style

        # Build this control's message callback mapping.
        self._message_callbacks = {}
        for name, callback in list(kwargs.items()):
            message = self._message_names[name]
            self._message_callbacks[message] = callback

        # Get a control ID and register this control with its parent.
        self._id = self._get_next_id()
        self._handle = None
        self._parent.add_control(self)


    #-----------------------------------------------------------------------

    def _get_handle(self):
        if self._handle is not None:
            return self._handle
        self._handle = win32gui.GetDlgItem(self._parent.hwnd, self.id)
        return self._handle

    handle = property(lambda self: self._get_handle())


    #-----------------------------------------------------------------------

    def _get_text(self):
        return win32gui.GetWindowText(self.handle)

    def _set_text(self, text):
        return win32gui.SetWindowText(self.handle, text)

    text = property(
                    fget=lambda self: self._get_text(),
                    fset=lambda self, text: self._set_text(text),
                   )


    #-----------------------------------------------------------------------

    id                 = property(lambda self: self._id)
    message_callbacks  = property(lambda self: self._message_callbacks)


    #-----------------------------------------------------------------------

    def calculate_size(self, width, height):
        return self._size(width, height)

    def template_entry(self, width, height):
        entry = [self._flavor, self._text, self._id,
                 self.calculate_size(width, height), self._style]
        return entry

    def enable(self):
        style = win32gui.GetWindowLong(self.handle, win32con.GWL_STYLE)
        style &= (~win32con.WS_DISABLED)
        win32gui.SetWindowLong(self.handle, win32con.GWL_STYLE, style)
        win32gui.SendMessage(self.handle, win32con.WM_ENABLE, 1, 0)

    def disable(self):
        style = win32gui.GetWindowLong(self.handle, win32con.GWL_STYLE)
        style |= win32con.WS_DISABLED
        win32gui.SetWindowLong(self.handle, win32con.GWL_STYLE, style)
        win32gui.SendMessage(self.handle, win32con.WM_ENABLE, 0, 0)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements an edit control which can be used as a Python stream
for text-based output.  It can for example be assigned to standard
output and standard error to display Python print statements and
exception trace back.

"""


#---------------------------------------------------------------------------

import sys
import win32con
import win32gui
import logging

from dragonfly.windows.control_base import ControlBase


#---------------------------------------------------------------------------

class OutputLogHandler(logging.Handler):

    def __init__(self, control):
        self._control = control
        logging.Handler.__init__(self)

    def emit(self, record):
        message = self.format(record)
        message = "\r\n".join(message.splitlines()) + "\r\n"
        self._control.write(message)


#---------------------------------------------------------------------------

class OutputText(ControlBase):

    def __init__(self, parent, size, **kwargs):
        flavor = "EDIT"
        text = ""
        style = ( win32con.WS_CHILD
                | win32con.WS_VISIBLE
                | win32con.ES_LEFT
                | win32con.ES_MULTILINE
                | win32con.ES_AUTOVSCROLL
                | win32con.WS_TABSTOP
                | win32con.WS_BORDER
                | win32con.ES_READONLY
                | win32con.WS_VSCROLL)
        ControlBase.__init__(self, parent, flavor, text, size, style, **kwargs)

    def set_as_output(self):
        self._old_stdout = sys.stdout
        self._old_stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self

    def set_as_log_handler(self, name="", level=logging.DEBUG):
        handler = OutputLogHandler(self)
        handler.setLevel(level)
        formatter = logging.Formatter("%(asctime)s %(name)s (%(levelname)s):"
                                        " %(message)s")
        handler.setFormatter(formatter)
        logging.getLogger(name).addHandler(handler)

    def write(self, data):
        length = win32gui.GetWindowTextLength(self.handle)
        win32gui.SendMessage(self.handle, win32con.EM_SETSEL, length, length)
        win32gui.SendMessage(self.handle, win32con.EM_REPLACESEL, False, data)

    def flush(self):
        pass




import time
from collections  import defaultdict
from ...grammar.rule_base import Rule
from ...grammar.context import AppContext
from ...windows.window import Window
from ...actions.action_base import ActionError


#===========================================================================

class StateBase(object):

    toplevel = True

    def __init__(self, name=None):
        self._name = name
        self._transitions = []
        self._include_states = []
        self._command_chain_rules = {}
        self._recursing_transitions = False

    def __str__(self):
        return "%s(name=%s)" % (self.__class__.__name__, self._name)

    #-----------------------------------------------------------------------

    def _get_name(self):
        return self._name
    def _set_name(self, name):
        self._name = name
    name = property(_get_name, _set_name)

    @property
    def transitions(self):
        if self._recursing_transitions:
            return ()
        self._recursing_transitions = True
        try:
            transitions = self._transitions
            for state in self._include_states:
                for transition in state.transitions:
                    transitions.append(transition.copy_with_prestate(self))
        finally:
            self._recursing_transitions = False
        return transitions

    def add_transitions(self, *transitions):
        for transition in transitions:
            if transition not in self._transitions:
                self._transitions.append(transition)

    def include_states(self, *states):
        for state in states:
            if state not in self._include_states:
                self._include_states.append(state)

    #-----------------------------------------------------------------------

    def get_command_chain_rule(self, length, toplevel=False):
        try:
            rule = self._command_chain_rules[length]
        except KeyError:
            chain_element = self._build_command_chain(length)
            global unique; unique += 1
            name = "_%s_%d_%03d" % (self, length, unique-1)
            if toplevel:
                rule = ToplevelRule(element=chain_element, name=name)
            else:
                rule = Rule(element=chain_element, name=name)
            self._command_chain_rules[length] = rule
        return rule

    def _build_command_chain(self, length):
        if length < 1:
            return None
        elif length == 1:
            poststate_ref_pairs = self._build_poststate_ref_pairs()
            alternatives = [ref for pstate, ref in poststate_ref_pairs]
            return Alternative(alternatives)
        else:
            poststate_ref_pairs = self._build_poststate_ref_pairs()
            alternatives = []
            for pstate, ref in poststate_ref_pairs:
                tail = pstate.get_command_chain_rule(length - 1)
                pstate_element = Sequence([ref, Optional(RuleRef(tail))])
                alternatives.append(pstate_element)
            return Alternative(alternatives)

    def _build_poststate_ref_pairs(self):
        poststate_mapping = defaultdict(list)
        poststate_id_mapping = {}
        for command_set in self.command_sets:
            pairs = command_set.get_poststate_commands_pairs()
            for poststate, commands in pairs:
                if not poststate:
                    poststate = self
                poststate_mapping[id(poststate)].extend(commands)
                poststate_id_mapping[id(poststate)] = poststate
        poststate_ref_pairs = []
        for poststate_id, commands in poststate_mapping.items():
            poststate = poststate_id_mapping[poststate_id]
            command_refs = [CommandRef(cmd, self) for cmd in commands]
            command_element = Alternative(command_refs)
            global unique; unique += 1
            name = "_%s_%s_%03d" % (self, poststate, unique-1)
            command_rule = Rule(element=command_element, name=name)
            poststate_ref_pairs.append((poststate, RuleRef(command_rule)))
        return poststate_ref_pairs

    #-----------------------------------------------------------------------

    def is_active(self):
        return False

    def is_active_toplevel(self):
        return self.is_active()


#---------------------------------------------------------------------------

class GlobalState(StateBase):

    def __init__(self):
        StateBase.__init__(self)

    def is_active(self):
        return True


#---------------------------------------------------------------------------

class StopState(StateBase):

    toplevel = False

    def __init__(self):
        StateBase.__init__(self)

    def is_active(self):
        return False


#---------------------------------------------------------------------------

class HiddenState(StateBase):

    toplevel = False

    def __init__(self):
        StateBase.__init__(self)

    def is_active(self):
        return True

    def is_active_toplevel(self):
        return False


#---------------------------------------------------------------------------

class WindowState(StateBase):

    def __init__(self, executable=None, title=None):
        StateBase.__init__(self)
        self._executable  = executable
        self._title       = title
        self._context     = AppContext(executable=executable, title=title)

    def __str__(self):
        args = []
        if self._executable:  args.append(self._executable)
        if self._title:       args.append(self._title)
        args = ",".join(args)
        return "%s(%s)" % (self.__class__.__name__, args)

    def is_active(self):
        window = Window.get_foreground()
        return self._context.matches(
                                     executable=window.executable,
                                     title=window.title,
                                     handle=window.handle,
                                    )


#===========================================================================

class Transition(object):

    def __init__(self, command, prestate, poststate):
        self._command    = command
        self._prestate   = prestate
        self._poststate  = poststate
        self._timeout    = 1.0
        self._interval   = 0.1

    @property
    def prestate(self):
        return self._prestate

    @property
    def poststate(self):
        return self._poststate

    @property
    def command(self):
        return self._command

    def copy_with_prestate(self, prestate):
        clone = Transition(self._command, prestate, self._poststate)
        return clone

    def bind(self, node):
        return BoundTransition(self, node)

    def execute(self, node):
        timeout_time = time.time() + self._timeout
        while not self.prestate.is_active():
            if time.time() > timeout_time:
                raise ActionError("prestate not active: %s" % self.prestate)
            time.sleep(self._interval)
        self._command.execute(node)




#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Parser elements specific to the command family input forms
============================================================================

"""

import string
from ...parser import (Parser, ParserElementBase,
                       Sequence, Alternative, Repetition, Optional,
                       String, Choice, CharacterSeries, Whitespace,
                       QuotedString, UnsignedInteger, Integer, Float)


#---------------------------------------------------------------------------

class Identifier(object):

    def __init__(self, name):
        self.name = name    

    def __str__(self):
        return "Identifier(%r)" % self.name

    def __repr__(self):
        return "Identifier(%r)" % self.name

    def __ne__(self, other):
        return not (self == other)

    def __eq__(self, other):
        if isinstance(other, Identifier):
            return self.name == other.name
        else:
            return NotImplemented


class Argument(object):

    def __init__(self, name=None, value=None):
        self.name = name    
        self.value = value

    def __str__(self):
        return "Argument(%r, %r)" % (self.name, self.value)

    def __ne__(self, other):
        return not (self == other)

    def __eq__(self, other):
        if isinstance(other, Argument):
            return self.name == other.name and self.value == other.value
        else:
            return NotImplemented


class Call(object):

    def __init__(self, function, arguments):
        self.function = function
        self.arguments = arguments

    def __str__(self):
        args = [str(a) for a in self.arguments]
        return "Call(f=%r, args=(%s))" % (self.function, ", ".join(args))

    def __repr__(self):
        return self.__str__()

    def __ne__(self, other):
        return not (self == other)

    def __eq__(self, other):
        if isinstance(other, Call):
            if self.function != other.function:
                return False
            if len(self.arguments) != len(other.arguments):
                return False
            for mine, theirs in zip(self.arguments, other.arguments):
                if mine != theirs:
                    return False
            return True
        else:
            return NotImplemented


#---------------------------------------------------------------------------

class IdentifierElement(CharacterSeries):
    """
        Parser element for identifiers.

        Usage examples:
        >>> from dragonfly.parser import Parser
        >>> parser = Parser(IdentifierElement())
        >>> def test_identifier_element(input):
        ...     output = parser.parse(input)
        ...     if not output: return False
        ...     assert isinstance(output, Identifier)
        ...     assert output.name == input
        ...     return True
        >>> assert test_identifier_element("foo_bar")
        >>> assert test_identifier_element("foo.bar")
        >>> assert test_identifier_element("foo.bar_baz")
        >>> assert not test_identifier_element("invalid/character")
        >>> assert not test_identifier_element("invalid character")

    """

    valid_chars = "".join((
                           string.uppercase,
                           string.lowercase,
                           string.digits,
                           "_.",
                         ))

    def __init__(self, name=None):
        CharacterSeries.__init__(self, set=self.valid_chars, name=name)

    def value(self, node):
        name = CharacterSeries.value(self, node)
        return Identifier(name)


class ValueElement(Alternative):
    """
        Parser element for values.

        Usage examples:
        >>> from six import string_types
        >>> from dragonfly.parser import Parser
        >>> parser = Parser(ValueElement())
        >>> def test_value_element(input, identifier):
        ...     output = parser.parse(input)
        ...     if not output: return False
        ...     if identifier:
        ...         assert isinstance(output, Identifier)
        ...         assert output.name == input
        ...     else:
        ...         assert isinstance(output, string_types)
        ...         input = input.decode("string_escape")  # Interpret escaped chars.
        ...         assert output == input[1:-1]           # Strip off quotation marks.
        ...     return True
        >>> assert test_value_element("foo_bar", True)
        >>> assert test_value_element("foo.bar", True)
        >>> assert test_value_element("'foo_bar'", False)
        >>> assert test_value_element('"foo_bar"', False)
        >>> assert test_value_element(r'"Hello \\"world\\"!"', False)

    """

    def __init__(self, name=None):
        children = (
                    Integer(),
                    Float(),
                    QuotedString(),
                    IdentifierElement(),
                   )
        Alternative.__init__(self, children, name=name)


class ArgumentElement(Sequence):
    """
        Parser element for call arguments.

        Usage examples:
        >>> from six import string_types
        >>> from dragonfly.parser import Parser
        >>> parser = Parser(ArgumentElement())
        >>> def test_argument_element(input, name, value):
        ...     output = parser.parse(input)
        ...     if not output: return False
        ...     assert isinstance(output.name, string_types) or output.name is None
        ...     assert output.name == name
        ...     assert isinstance(output.value, (string_types, Identifier))
        ...     assert output.value == value
        ...     return True
        >>> assert test_argument_element("foo_bar",     None,     Identifier("foo_bar"))
        >>> assert test_argument_element('"foo"',       None,     "foo")
        >>> assert test_argument_element("foo=bar",     "foo",    Identifier("bar"))
        >>> assert test_argument_element("foo='bar'",   "foo",    "bar")
        >>> assert test_argument_element("foo ='bar'",  "foo",    "bar")
        >>> assert test_argument_element("foo= 'bar'",  "foo",    "bar")
        >>> assert test_argument_element("foo = 'bar'", "foo",    "bar")
        >>> assert test_argument_element("foo=\t'bar'", "foo",    "bar")
        >>> assert not test_argument_element("=foo",    None,     None)
        >>> assert not test_argument_element("foo=",    None,     None)
        >>> assert not test_argument_element("'foo'=bar", None,   None)

    """

    def __init__(self, name=None):
        ws = Optional(Whitespace())
        children = (
                    Optional(
                             Sequence([
                                       IdentifierElement(name="argid"), ws,
                                       String("="), ws,
                                     ])
                            ),
                    ValueElement(),
                   )
        Sequence.__init__(self, children, name=name)

    def value(self, node):
        name_node = node.get_child(name="argid",
                                   actor_type=IdentifierElement)
        if name_node:  name = name_node.value().name
        else:          name = None
        value = node.children[1].value()
        return Argument(name, value)


#---------------------------------------------------------------------------

class CallElement(Sequence):
    """
        Parser element for function call expressions.

        Usage examples:
        >>> from logging import getLogger
        >>> from dragonfly.parser import Parser
        >>> p = Parser(CallElement())
        >>> assert p.parse("foo()")           == Call("foo", [])
        >>> assert p.parse("foo(bar)")        == Call("foo", [Argument(value=Identifier("bar"))])
        >>> assert p.parse("foo(bar, baz)")   == Call("foo", [Argument(value=Identifier("bar")), \
                                                              Argument(value=Identifier("baz"))])
        >>> assert p.parse("foo(bar, 'baz')") == Call("foo", [Argument(value=Identifier("bar")), \
                                                              Argument(value="baz")])
        >>> assert p.parse("foo ( bar , \t'baz' )") == Call("foo", [Argument(value=Identifier("bar")), \
                                                                    Argument(value="baz")])

    """

    def __init__(self, name=None):
        ws = Optional(Whitespace())
        function = IdentifierElement(name="function")
        argument = ArgumentElement(name="arg")
        argument_list = Optional(Sequence([
                                  argument,
                                  Repetition(
                                             Sequence([
                                                       ws, String(","),
                                                       ws, argument,
                                                     ]),
                                             min=0,
                                             max=None,
                                            ),
                                ]))
        children = (
                    function,
                    ws,
                    String("("),
                    ws,
                    argument_list,
                    ws,
                    String(")"),
                   )
        Sequence.__init__(self, children, name)

    def value(self, node):
        function = node.get_child(name="function").value().name
        arguments = [n.value() for n in node.get_children(name="arg")]
        return Call(function, arguments)


#---------------------------------------------------------------------------

class _ActionGroup(object):
    def __init__(self, items, factor=1):
        self.items = items
        self.factor = factor
    def build_tree(self):
        if len(self.items) == 1 and self.factor == 1:
            if isinstance(self.items[0], self.__class__):
                return self.items[0].build_tree()
            else:
                return self.items[0]
        branches = []
        for item in self.items:
            if isinstance(item, self.__class__):
                branch = item.build_tree()
                if isinstance(branch, tuple) and branch[1] == 1:
                    branches.extend(branch[0])
                else:
                    branches.append(branch)
            else:
                branches.append(item)
        return (branches, self.factor)

class ActionElement(Sequence):
    """
        Parser element for action expressions.

        Usage examples:
        >>> p = Parser(ActionElement())
        >>> from pprint import PrettyPrinter
        >>> pretty = lambda input: PrettyPrinter().pprint(p.parse(input))
        >>> pretty("foo()")
        Call(f='foo', args=())
        >>> pretty("(\t (foo()\t)* 1  )")
        Call(f='foo', args=())
        >>> pretty("foo() bar()")
        ([Call(f='foo', args=()), Call(f='bar', args=())], 1)
        >>> pretty("(foo()( bar())) baz()")
        ([Call(f='foo', args=()), Call(f='bar', args=()), Call(f='baz', args=())], 1)
        >>> pretty("foo() * 9 bar()")
        ([([Call(f='foo', args=())], 9), Call(f='bar', args=())], 1)
        >>> pretty("foo() * 'integer' bar()")
        ([([Call(f='foo', args=())], 'integer'), Call(f='bar', args=())], 1)
        >>> pretty("(foo() bar())*3 baz()")
        ([([Call(f='foo', args=()), Call(f='bar', args=())], 3),
          Call(f='baz', args=())],
         1)
        >>> pretty("((foo() bar())*3 (foo() bar())*'count')*7")
        ([([Call(f='foo', args=()), Call(f='bar', args=())], 3),
          ([Call(f='foo', args=()), Call(f='bar', args=())], 'count')],
         7)
        >>> pretty("(Key("") Key("")) * 7 Key("") * 3")
        ([([Call(f='Key', args=()), Call(f='Key', args=())], 7),
          ([Call(f='Key', args=())], 3)],
         1)

    """

    #-----------------------------------------------------------------------

    class MulExpression(Sequence):
        def __init__(self, unit_mul, ws):
            mul_factor = Alternative((
                                      UnsignedInteger(),
                                      QuotedString(),
                                    ))
            Sequence.__init__(self,
                              (
                               unit_mul,
                               Repetition(
                                          Sequence((
                                                    ws, String("*"),
                                                    ws, mul_factor,
                                                  )),
                                          min=1,
                                          #max=8,
                                         )
                             ))
        def value(self, node):
            unit_mul_value = node.children[0].value()
            factors = []
            for repetition in node.children[1].value():
                factors.append(repetition[3])
            result = unit_mul_value
            for f in factors:
                result = _ActionGroup([result], f)
            return result

    class AddExpression(Sequence):
        def __init__(self, unit_add, ws):
            Sequence.__init__(self,
                              (
                               unit_add,
                               Repetition(
                                          Sequence((
                                                    ws, unit_add,
                                                  )),
                                          min=0,
                                          #max=8,
                                         )
                             ))
        def value(self, node):
            unit_add_values = [node.children[0].value()]
            for repetition in node.children[1].value():
                unit_add_values.append(repetition[1])
            return _ActionGroup(unit_add_values)

    class GrpExpression(Sequence):
        def __init__(self, expr_add, ws):
            Sequence.__init__(self,
                              (
                               String("("),
                               ws, expr_add,
                               ws, String(")"),
                             ))
        def value(self, node):
            expr_add_value = node.children[2].value()
            return _ActionGroup([expr_add_value])

    single_call = CallElement()
    ws = Optional(Whitespace())
    unit_add = Alternative()
    unit_mul = Alternative()
    expr_mul = MulExpression(unit_mul, ws)
    expr_add = AddExpression(unit_add, ws)
    expr_grp = GrpExpression(expr_add, ws)
    unit_add.child_list.extend((single_call, expr_grp, expr_mul))
    unit_mul.child_list.extend((single_call, expr_grp))

    #-----------------------------------------------------------------------

    def __init__(self, name=None):
        children = (
                    self.expr_add,
                   )
        Sequence.__init__(self, children, name)

    def value(self, node):
        group = node.children[0].value()
        tree = group.build_tree()
        return tree


#---------------------------------------------------------------------------

class _ContextGroup(object):
    def __init__(self, operator, before, after=None):
        self.operator  = operator
        self.before    = before
        self.after     = after
    def build_tree(self):
        if isinstance(self.before, self.__class__):
            before = self.before.build_tree()
        else:
            before = self.before
        if self.after:
            if isinstance(self.after, self.__class__):
                after = self.after.build_tree()
            else:
                after = self.after
            return (self.operator, after, before)
        else:
            return (self.operator, before)

class ContextElement(Sequence):
    """
        Parser element for context expressions.

        Usage examples:
        >>> p = Parser(ContextElement())
        >>> from pprint import PrettyPrinter
        >>> pretty = lambda input: PrettyPrinter().pprint(p.parse(input))
        >>> pretty("foo()")
        Call(f='foo', args=())
        >>> pretty("foo() & bar()")
        ('&', Call(f='foo', args=()), Call(f='bar', args=()))
        >>> pretty("foo() & bar() | baz()")
        ('|',
         ('&', Call(f='foo', args=()), Call(f='bar', args=())),
         Call(f='baz', args=()))
        >>> pretty("foo() & (bar() | baz())")
        ('&',
         Call(f='foo', args=()),
         ('|', Call(f='bar', args=()), Call(f='baz', args=())))
        >>> pretty("!foo()")
        ('!', Call(f='foo', args=()))
        >>> pretty("foo() | !bar()")
        ('|', Call(f='foo', args=()), ('!', Call(f='bar', args=())))
        >>> pretty("!foo() & !(bar() | baz())")
        ('&',
         ('!', Call(f='foo', args=())),
         ('!', ('|', Call(f='bar', args=()), Call(f='baz', args=()))))

    """

    #-----------------------------------------------------------------------

    class SeriesExpression(Sequence):
        def __init__(self, unit, operator, ws):
            Sequence.__init__(self,
                              (
                               unit,
                               Repetition(
                                          Sequence((
                                                    ws, operator,
                                                    ws, unit,
                                                  )),
                                          min=0,
                                         )
                             ))
        def value(self, node):
            values = [node.children[0].value()]
            for repetition in node.children[1].value():
                operator_value  = repetition[1]
                unit_value      = repetition[3]
                values.extend((operator_value, unit_value))
            if len(values) == 1:
                return values[0]
            values.reverse()
            after     = values.pop()
            operator  = values.pop()
            before    = values.pop()
            group = _ContextGroup(operator, before, after)
            while values:
                operator  = values.pop()
                before    = values.pop()
                group     = _ContextGroup(operator, before, group)
            return group

    class NotExpression(Sequence):
        def __init__(self, unit_not, ws):
            Sequence.__init__(self, (String("!"), ws, unit_not))
        def value(self, node):
            operator  = node.children[0].value()
            value     = node.children[2].value()
            group     = _ContextGroup(operator, value)
            return group

    class GrpExpression(Sequence):
        def __init__(self, expr_series, ws):
            Sequence.__init__(self,
                              (
                               String("("),
                               ws, expr_series,
                               ws, String(")"),
                             ))
        def value(self, node):
            series_value = node.children[2].value()
            return series_value

    single_call  = CallElement()
    ws           = Optional(Whitespace())
    operator     = Alternative((String("&"), String("|")))
    unit_series  = Alternative()
    unit_not     = Alternative()
    expr_series  = SeriesExpression(unit_series, operator, ws)
    expr_grp     = GrpExpression(expr_series, ws)
    expr_not     = NotExpression(unit_not, ws)
    unit_not.child_list.extend((single_call, expr_grp))
    unit_series.child_list.extend((single_call, expr_not, expr_grp))

    #-----------------------------------------------------------------------

    def __init__(self, name=None):
        children = (
                    self.expr_series,
                   )
        Sequence.__init__(self, children, name)

    def value(self, node):
        root = node.children[0].value()
        if isinstance(root, _ContextGroup):
            tree = root.build_tree()
        else:
            tree = root
        return tree



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""This file implements a Win32 keyboard interface using sendinput."""


import time
from six import text_type, PY2

import win32con

from ctypes import windll, c_char, c_wchar
from dragonfly.actions.sendinput import (KeyboardInput, make_input_array,
                                         send_input_array)


class Typeable(object):
    """Container for keypress events."""

    __slots__ = ("_code", "_modifiers", "_name", "_is_text")

    def __init__(self, code, modifiers=(), name=None, is_text=False):
        """Set keypress information."""
        self._code = code
        self._modifiers = modifiers
        self._name = name
        self._is_text = is_text

    def __str__(self):
        """Return information useful for debugging."""
        return ("%s(%s)" % (self.__class__.__name__, self._name) +
                repr(self.events()))

    def on_events(self, timeout=0):
        """Return events for pressing this key down."""
        if self._is_text:
            events = [(self._code, True, timeout, True)]
        else:
            events = [(m, True, 0) for m in self._modifiers]
            events.append((self._code, True, timeout))
        return events

    def off_events(self, timeout=0):
        """Return events for releasing this key."""
        if self._is_text:
            events = [(self._code, False, timeout, True)]
        else:
            events = [(m, False, 0) for m in self._modifiers]
            events.append((self._code, False, timeout))
            events.reverse()
        return events

    def events(self, timeout=0):
        """Return events for pressing and then releasing this key."""
        if self._is_text:
            events = [(self._code, True, timeout, True),
                      (self._code, False, timeout, True)]
        else:
            events = [(self._code, True, 0), (self._code, False, timeout)]
            for m in self._modifiers[-1::-1]:
                events.insert(0, (m, True, 0))
                events.append((m, False, 0))
        return events


class Keyboard(object):
    """Static class wrapper around SendInput."""

    shift_code = win32con.VK_SHIFT
    ctrl_code = win32con.VK_CONTROL
    alt_code = win32con.VK_MENU

    @classmethod
    def send_keyboard_events(cls, events):
        """
        Send a sequence of keyboard events.

        Positional arguments:
        events -- a sequence of tuples of the form
            (keycode, down, timeout), where
                keycode (int): Win32 Virtual Key code.
                down (boolean): True means the key will be pressed down,
                    False means the key will be released.
                timeout (int): number of seconds to sleep after
                    the keyboard event.
            or
            (character, down, timeout, is_text), where
                character (str): Unicode character.
                down (boolean): True means the key will be pressed down,
                    False means the key will be released.
                timeout (int): number of seconds to sleep after
                    the keyboard event.
                is_text (boolean): True means that the keypress is targeted
                    at a window or control that accepts Unicode text.
        """
        items = []
        for event in events:
            if len(event) == 3:
                keycode, down, timeout = event
                input_structure = KeyboardInput(keycode, down)
            elif len(event) == 4 and event[3]:
                character, down, timeout = event[:3]
                input_structure = KeyboardInput(0, down, scancode=character)
            items.append(input_structure)
            if timeout:
                array = make_input_array(items)
                items = []
                send_input_array(array)
                time.sleep(timeout)
        if items:
            array = make_input_array(items)
            send_input_array(array)
            if timeout: time.sleep(timeout)

    @classmethod
    def _get_initial_keycode(cls, char):
        layout = windll.user32.GetKeyboardLayout(0)
        if isinstance(char, str) and PY2:
            code = windll.user32.VkKeyScanExA(c_char(char), layout)
        elif isinstance(char, text_type):  # unicode for PY2, str for PY3
            code = windll.user32.VkKeyScanExW(c_wchar(char), layout)
        else:
            code = -1
        if code == -1:
            raise ValueError("Unknown char: %r" % char)
        return code

    @classmethod
    def xget_virtual_keycode(cls, char):
        code = cls._get_initial_keycode(char)

        # Construct a list of the virtual key code and modifiers.
        codes = [code & 0x00ff]
        if code & 0x0100: codes.append(cls.shift_code)
        if code & 0x0200: codes.append(cls.ctrl_code)
        if code & 0x0400: codes.append(cls.alt_code)
        return codes

    @classmethod
    def get_keycode_and_modifiers(cls, char):
        code = cls._get_initial_keycode(char)

        # Construct a list of the virtual key code and modifiers.
        modifiers = []
        if code & 0x0100: modifiers.append(cls.shift_code)
        if code & 0x0200: modifiers.append(cls.ctrl_code)
        if code & 0x0400: modifiers.append(cls.alt_code)
        code &= 0x00ff
        return code, modifiers

    @classmethod
    def get_typeable(cls, char, is_text=False):
        if is_text:
            return Typeable(char, is_text=True)
        code, modifiers = cls.get_keycode_and_modifiers(char)
        return Typeable(code, modifiers)


keyboard = Keyboard()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Multiplexing interface to a timer
============================================================================

"""

import time
import logging


#---------------------------------------------------------------------------

class Timer(object):

    _log = logging.getLogger("engine.timer")

    def __init__(self, function, interval, manager):
        self.function = function
        self.interval = interval
        self.manager = manager
        self.active = False
        self.next_time = None
        self.start()

    def start(self):
        if self.active:
            return
        self.manager.add_timer(self)
        self.active = True
        self.next_time = time.clock() + self.interval

    def stop(self):
        if not self.active:
            return
        self.manager.remove_timer(self)
        self.active = False
        self.next_time = None

    def call(self):
        self.next_time += self.interval
        try:
            self.function()
        except Exception as e:
            self._log.exception("Exception during timer callback: %s" % (e,))


class TimerManagerBase(object):
    """
    """

    _log = logging.getLogger("engine.timer")

    def __init__(self, interval, engine):
        """
        """
        self.interval = interval
        self.engine = engine
        self.timers = []

    def add_timer(self, timer):
        self.timers.append(timer)
        if len(self.timers) == 1:
            self._activate_main_callback(self.main_callback,
                                         self.interval)

    def remove_timer(self, timer):
        try:
            self.timers.remove(timer)
        except Exception as e:
            self._log.exception("Failed to remove timer: %s" % e)
            return
        if len(self.timers) == 0:
            self._deactivate_main_callback()

    def main_callback(self):
        now = time.clock()
        for c in self.timers:
            if c.next_time < now:
                try:
                    c.call()
                except Exception as e:
                    self._log.exception("Exception occurred during"
                                        " timer callback: %s" % (e,))

    def _activate_main_callback(self, callback, msec):
        raise NotImplementedError("_activate_main_callback() not implemented for"
                                  " %s class." % self.__class__.__name__)

    def _deactivate_main_callback(self):
        raise NotImplementedError("_deactivate_main_callback() not implemented for"
                                  " %s class." % self.__class__.__name__)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
FocusWindow action
============================================================================

"""


import win32con
from .action_base      import ActionBase, ActionError
from ..windows.window  import Window


#---------------------------------------------------------------------------

class FocusWindow(ActionBase):
    """
        Bring a window to the foreground action.

        Constructor arguments:
         - *executable* (*str*) -- part of the filename of the
           application's executable to which the target window belongs;
           not case sensitive.
         - *title* (*str*) -- part of the title of the target window;
           not case sensitive.

        This action searches all visible windows for a window which 
        matches the given parameters.

    """

    def __init__(self, executable=None, title=None):
        if executable:  self.executable = executable.lower()
        else:           self.executable = None
        if title:       self.title = title.lower()
        else:           self.title = None
        ActionBase.__init__(self)

        arguments = []
        if executable:  arguments.append("executable=%r" % executable)
        if title:       arguments.append("title=%r" % title)
        self._str = ", ".join(arguments)

    def _execute(self, data=None):
        executable = self.executable
        title = self.title
        if data and isinstance(data, dict):
            if executable:  executable = (executable % data).lower()
            if title:       title = (title % data).lower()

        windows = Window.get_all_windows()
        for window in windows:
            if not window.is_visible:
                continue
            if (window.executable.endswith("natspeak.exe")
                and window.classname == "#32770"
                and window.get_position().dy < 50):
                # If a window matches the above, it is very probably
                #  the results-box of DNS.  We ignore this because
                #  its title is the words of the last recognition,
                #  which will often interfere with a search for
                #  a window with a spoken title.
                continue

            if executable:
                if window.executable.lower().find(executable) == -1:
                    continue
            if title:
                if window.title.lower().find(title) == -1:
                    continue
            window.set_foreground()
            return
        raise ActionError("Failed to find window (%s)."  % self._str)



"""
GrammarWrapper class for CMU Pocket Sphinx engine
"""
import functools
from six import text_type

from jsgf import RuleRef, map_expansion, Literal, find_expansion, filter_expansion
from jsgf.ext import SequenceRule, DictationGrammar, only_dictation_in_expansion, \
    dictation_in_expansion

import dragonfly.grammar.state as state_
from dragonfly import Grammar
from .dragonfly2jsgf import LinkedRule


@functools.total_ordering
class ProcessingState(object):
    """
    State class used to process matching rules in a GrammarWrapper.

    ProcessingState objects can be compared to find the best state to process.
    """
    def __init__(self, wrapper, speech, use_linked_rules=False):
        """
        :type wrapper: GrammarWrapper
        :type speech: str
        :type use_linked_rules: bool
        """
        self._wrapper = wrapper
        self._post_collection_tasks = []
        self.speech = speech

        # This is used when mimicking speech as words.
        self.use_linked_rules = use_linked_rules
        self.in_progress_rules = []
        self.complete_rules = []
        self.speech_is_dictation = self.wrapper.engine.recognising_dictation

        # Collect matching rules into this state's rule lists.
        self.wrapper.collect_matching_rules(self)

    @property
    def wrapper(self):
        """
        The GrammarWrapper this state object was created for.
        :rtype: GrammarWrapper
        """
        return self._wrapper

    @property
    def is_processable(self):
        """
        Whether this state can be processed by its GrammarWrapper.
        """
        return bool(self.in_progress_rules or self.complete_rules)

    @property
    def complete_sequence_rules(self):
        return [x for x in self.complete_rules if isinstance(x, SequenceRule)]

    @property
    def repeatable_complete_sequence_rules(self):
        return [x for x in self.complete_sequence_rules if x.can_repeat]

    @property
    def complete_normal_rules(self):
        return [x for x in self.complete_rules if not isinstance(x, SequenceRule)]

    @property
    def matched_rules(self):
        return self.in_progress_rules + self.complete_rules

    def process(self, timed_out=False, notify_only=False):
        """
        Process this state object by calling the wrapper's process method.
        """
        self.wrapper.process(self, timed_out, notify_only)

    def add_post_collection_task(self, task):
        """
        Add a post collection task to be executed when `do_post_collection_tasks` is
        called.
        :param task: callable object
        :return:
        """
        if not callable(task):
            raise TypeError("%s is not a callable object" % task)
        self._post_collection_tasks.append(task)

    def do_post_collection_tasks(self):
        # Execute each task in the order they were added.
        while self._post_collection_tasks:
            self._post_collection_tasks.pop(0)()

    def __lt__(self, other):
        return (
            # Favour processable states.
            not self.is_processable and other.is_processable

            # Favour a specific grammar context over no context.
            or not self.wrapper.grammar._context and other.wrapper.grammar._context
        )

    def __eq__(self, other):
        return (self.speech == other.speech and self.wrapper == other.wrapper
                and self.in_progress_rules == other.in_progress_rules
                and self.complete_rules == other.complete_rules)

    def __hash__(self):
        return hash(self.wrapper.grammar.name)


class GrammarWrapper(object):
    def __init__(self, grammar, engine):
        """
        :type grammar: Grammar
        :type engine: SphinxEngine
        """
        self.grammar = grammar
        self.engine = engine

        # Compile the grammar into a JSGF grammar and add those rules into
        # a JSGF DictationGrammar.
        self._jsgf_grammar = engine.compiler.compile_grammar(grammar)
        self._jsgf_grammar.language_name = engine.language
        self._dictation_grammar = DictationGrammar(self._jsgf_grammar.rules)
        self._dictation_grammar.language_name = engine.language

        # Set internal variables
        self._in_progress_sequence_rules = []

        # Set the default search name based on whether only dictation can be matched
        grammar_has_jsgf = False
        for rule in self._dictation_grammar.match_rules:
            if not rule.active:
                continue  # skip disabled rules; they can't be matched
            if not only_dictation_in_expansion(rule.expansion):
                grammar_has_jsgf = True
                break

        if grammar_has_jsgf:
            self._default_search_name = self.grammar.name
        else:
            self._default_search_name = self.engine.dictation_search_name

        # Set the default value of search_name
        self._search_name = self._default_search_name

    @property
    def grammar_words(self):
        """
        Set of all words used in this grammar.

        :returns: set
        """
        words = []
        for rule in self._jsgf_grammar.rules:
            rule_literals = filter_expansion(
                rule.expansion, lambda x: isinstance(x, Literal) and x.text,
                shallow=True
            )
            for literal in rule_literals:
                words.extend(literal.text.split())

        # Return a set of words with no duplicates.
        return set(words)

    @property
    def dictation_grammar(self):
        return self._dictation_grammar

    @property
    def search_name(self):
        """
        The name of the Pocket Sphinx search that the engine should use to process
        speech for the grammar.
        :return: str
        """
        return self._search_name

    @property
    def default_search_name(self):
        """
        The name of the default Pocket Sphinx search that the engine should use to
        process speech for the grammar.

        This is not necessarily the current search that should be used for the
        grammar.
        """
        return self._default_search_name

    @property
    def grammar_active(self):
        """
        Whether the grammar is enabled and has any active rules.
        :rtype: bool
        """
        return self.grammar.enabled and any(self.grammar.active_rules)

    def reset_all_sequence_rules(self):
        """
        Resetting all SequenceRules, clear the in progress list and reset
        search_name.
        """
        for r in self._dictation_grammar.rules:
            if isinstance(r, SequenceRule):
                r.restart_sequence()

            # Also ensure all rules are now enabled again.
            r.enable()

        # Clear the in-progress list and set search_name back to default.
        self._in_progress_sequence_rules = []

        if self.search_name != self._default_search_name:
            # Unset the search used for in-progress rules. This won't unset the
            # dictation search.
            self.engine.unset_search(self.search_name)

            # Set search name back to default.
            self._search_name = self._default_search_name
            self.engine.set_grammar(self)

    def collect_matching_rules(self, state):
        """
        Collect rules matching state.speech (speech hypothesis) into the given
        state object's rule lists.
        The ProcessingState object can be processed using the GrammarWrapper
        `process` method.
        :type state: ProcessingState
        :rtype: ProcessingState
        """
        speech = state.speech
        if not ((speech or state.speech_is_dictation) or state.use_linked_rules):
            # Check if there are any match rules with dictation expansions
            dict_hyp_required = False
            for rule in self.dictation_grammar.match_rules:
                if dictation_in_expansion(rule.expansion):
                    dict_hyp_required = True
                    break

            if dict_hyp_required:
                dict_hyp = self.engine.get_dictation_hypothesis()
                if dict_hyp:
                    state.speech = dict_hyp
                    state.speech_is_dictation = True

        # Collect matching any in-progress SequenceRules
        if self._in_progress_sequence_rules:
            self._collect_in_progress_sequence_rules(state)

        # Collect matching normal rules
        self._collect_normal_rules(state)
        return state

    def process(self, state, timed_out=False, notify_only=False):
        """
        Process a ProcessingState object from `collect_matching_rules`.

        :param state: ProcessingState
        :param timed_out: used by recognition timeout thread
        :param notify_only: whether to only notify observers, not process
        """
        complete_seq = state.complete_sequence_rules
        if complete_seq:
            words_list = self._generate_words_list(complete_seq[0], True)
            self._process_complete_recognition(words_list, notify_only)

        elif state.in_progress_rules and not timed_out:
            self._notify_partial_recognition(state.in_progress_rules[0])

        elif state.complete_rules:
            words_list = self._generate_words_list(state.complete_rules[0], True)
            self._process_complete_recognition(words_list, notify_only)

    def process_begin(self, fg_window):
        """
        Start the dragonfly grammar processing.
        """
        self.grammar.process_begin(fg_window.executable, fg_window.title,
                                   fg_window.handle)

    def _process_results(self, words):
        """
        Start the dragonfly processing of the speech hypothesis.
        :param words: a sequence of (word, rule_id) 2-tuples (pairs)
        """
        self.engine.log.debug("Grammar %s: received recognition %r." %
                              (self.grammar.name, words))

        words_rules = tuple((text_type(w), r) for w, r in words)
        rule_ids = tuple(r for _, r in words_rules)
        words = tuple(w for w, r in words_rules)

        # Call the grammar's general process_recognition method, if present.
        func = getattr(self.grammar, "process_recognition", None)
        if func:
            if not func(words):
                return

        # Iterates through this grammar's rules, attempting to decode each.
        # If successful, call that rule's method for processing the recognition
        # and return.
        s = state_.State(words_rules, rule_ids, self.engine)
        for r in self.grammar.rules:
            if not r.active:
                continue
            s.initialize_decoding()

            # Iterate each result from decoding state 's' with grammar rule 'r'
            for _ in r.decode(s):
                if s.finished():
                    root = s.build_parse_tree()
                    r.process_recognition(root)
                    return

        self.engine.log.warning("Grammar %s: failed to decode recognition %r."
                                % (self.grammar.name, words))

    def _get_dragonfly_rule(self, rule):
        """
        Get the original dragonfly Rule given a JSGF Rule.
        :param rule: JSGF Rule
        :return: Rule
        """
        if isinstance(rule, LinkedRule):
            linked_rule = rule
        else:
            linked_rule = self._dictation_grammar.get_original_rule(rule)

        df_rule = linked_rule.df_rule
        return df_rule

    def _notify_partial_recognition(self, rule):
        """
        Internal method to notify the engine's next_rule_part observers of a
        partial recognition.
        :type rule: SequenceRule
        """
        self.engine.observer_manager.notify_next_rule_part(
            self._generate_words_list(rule, False)
        )

    def _process_complete_recognition(self, words_list, notify_only):
        """
        Internal method for processing complete recognitions.
        :type words_list: list
        :param notify_only: whether to only notify observers, not process.
        """
        # Notify recognition observers
        self.engine.observer_manager.notify_recognition(words_list)

        # Begin dragonfly processing if appropriate.
        # notify_only is used by the Sphinx engine's training mode.
        if not notify_only:
            try:
                self._process_results(words_list)
            except Exception as e:
                self.engine.log.error("%s: caught Exception while processing "
                                      "recognised words: %s" % (self, e))

        # Clear the in progress list and reset all sequence rules in the grammar.
        self.reset_all_sequence_rules()

        # Switch back to the relevant Pocket Sphinx search
        self.engine.set_grammar(self)

    def _generate_words_list(self, rule, complete_match):
        """
        Generate a words list compatible with dragonfly's processing classes.
        :param rule: JSGF Rule
        :param complete_match: whether all expansions in a SequenceRule must match
        :return: list
        """
        df_rule_id = self.grammar.rules.index(
            self._get_dragonfly_rule(rule)
        )

        if isinstance(rule, SequenceRule) or dictation_in_expansion(rule.expansion):
            words_list = []
            if isinstance(rule, SequenceRule):
                # Used by real speech input and engine.mimic_phrases.
                expansions = rule.expansion_sequence
            else:
                # LinkedRules cannot be partially matched, but this shouldn't
                # happen regardless, so log an error.
                if not rule.was_matched:
                    self.engine.log.error("LinkedRule %s returned an empty words "
                                          "list" % rule)
                    return []

                def matching(x):
                    # Note that Dictation is a subclass of jsgf.Literal so it
                    # doesn't need to be specified here.
                    if isinstance(x, (Literal, RuleRef)) and x.current_match:
                        # All ancestors must also have non-null match values.
                        p = x.parent
                        while p:
                            if not p.current_match:
                                return False
                            p = p.parent
                        return True

                # Initialise the list with the first matching expansion.
                first = find_expansion(rule.expansion, matching)
                expansions = [first]

                if not first:
                    self.engine.log.error("Matching rule %s had no matching "
                                          "expansions" % rule)
                    return []

                # Then add each matchable leaf afterwards. This does what
                # Expansion.matchable_leaves_after does, except that it only looks
                # at leaves in this tree, not referenced ones. This will not work
                # with repetition.
                shallow_leaves = first.root_expansion.collect_leaves(shallow=True)
                first_reached = False
                for leaf in shallow_leaves:
                    if leaf is first:
                        first_reached = True
                        continue
                    elif first_reached and (not first.mutually_exclusive_of(leaf)
                                            and matching(leaf)):
                        expansions.append(leaf)

            # Generate a words list using the match values for each expansion in
            # the list.
            for e in expansions:
                # Use rule IDs compatible with dragonfly's processing classes
                if only_dictation_in_expansion(e):
                    rule_id = 1000000  # dgndictation id
                else:
                    rule_id = df_rule_id

                # If the rule is not completely matched yet and complete_match is
                # False, then use the match values thus far. This should only happen
                # for SequenceRules.
                if not complete_match and e.current_match is None:
                    break

                # Get the words from the expansion's current match
                words = e.current_match.split()
                for word in words:
                    words_list.append((word, rule_id))

        # Non-sequential rules must match completely.
        # These rules shouldn't have any Dictation expansions.
        elif rule.was_matched:
            words_list = [(word, df_rule_id)
                          for word in rule.expansion.current_match.split()]
        else:
            words_list = []

        if not words_list:
            self.engine.log.debug("Rule %s had match '%s', but returned an empty "
                                  "words list"
                                  % (rule, rule.expansion.current_match))

        return words_list

    def _collect_normal_rules(self, state):
        """
        Internal method used to collect normal rules that can be spoken in one
        utterance that match a speech string.
        :type state: ProcessingState
        :rtype: ProcessingState
        """
        # No rules will match None or ""
        if not state.speech:
            return state

        # Find rules of active grammars that match the speech string
        if state.use_linked_rules:
            matching_rules = self._jsgf_grammar.find_matching_rules(state.speech)
            state.complete_rules.extend(matching_rules)
            return state
        else:
            matching_rules = self._dictation_grammar.find_matching_rules(
                state.speech, advance_sequence_rules=False)

        for rule in matching_rules:
            if isinstance(rule, SequenceRule):  # spoken in multiple utterances
                if rule.current_is_dictation_only and not state.speech_is_dictation:
                    # Get a dictation hypothesis
                    dict_hypothesis = self.engine.get_dictation_hypothesis()

                    # Note: if dict_hypothesis is None, then this recognition was
                    # probably invoked by mimic, so don't invalidate the match.
                    if dict_hypothesis:
                        rule.refuse_matches = False
                        rule.matches(dict_hypothesis)

                if rule.has_next_expansion:
                    # Add this rule to the in progress list and enqueue
                    # `rule.set_next()` as a post collection task.
                    state.add_post_collection_task(rule.set_next)
                    self._in_progress_sequence_rules.append(rule)
                    state.in_progress_rules.append(rule)
                else:
                    # The entire sequence has been matched. This rule could only
                    # have had one part to it.
                    state.complete_rules.append(rule)
            else:
                # This rule has been fully recognised.
                state.complete_rules.append(rule)
        return state

    def _collect_in_progress_sequence_rules(self, state):
        """
        Match against SequenceRules that are in progress and add them to the state
        object.
        :type state: ProcessingState
        """
        # No rules will match None or ""
        if not state.speech:
            state.add_post_collection_task(self.reset_all_sequence_rules)
            return state

        # Process the rules using a shallow copy of the in progress list because the
        # original list will be modified
        dict_hypothesis = self.engine.get_dictation_hypothesis()
        for rule in tuple(self._in_progress_sequence_rules):
            # Match the rule with the appropriate hypothesis string.
            if rule.current_is_dictation_only and dict_hypothesis:
                rule.matches(dict_hypothesis)
            else:
                rule.matches(state.speech)

            if not rule.was_matched:
                # Remove and reset the SequenceRule if it didn't match
                state.add_post_collection_task(rule.restart_sequence)
                self._in_progress_sequence_rules.remove(rule)
            elif rule.has_next_expansion:
                # By calling rule.matches, the speech value was stored as the
                # current expansion's current_match value.
                state.add_post_collection_task(rule.set_next)
                self._in_progress_sequence_rules.append(rule)
                state.in_progress_rules.append(rule)
            else:
                # SequenceRule has been completely matched.
                state.complete_rules.append(rule)

        # If the list still has rules, then it needs further processing.
        # Load a JSGF Pocket Sphinx search with just the rules in the list, or
        # switch to the dictation search if there are only dictation rules.
        if self._in_progress_sequence_rules:
            not_in_progress_rules = [r for r in self.dictation_grammar.match_rules
                                     if r not in self._in_progress_sequence_rules]

            # Disable rules in the grammar that aren't in the in-progress list
            for rule in not_in_progress_rules:
                rule.disable()

            # Recursively enable any rules referenced by an in-progress rule
            def enable_referenced(x):
                if isinstance(x, RuleRef):
                    x.rule.enable()

            for rule in self._in_progress_sequence_rules:
                map_expansion(rule.expansion, enable_referenced)

            # Set the search name for a new Pocket Sphinx search.
            # If we use a new name, Pocket Sphinx will keep the grammar's
            # original search around so we can switch back to it later without
            # recompiling and setting it again.
            self._search_name = "%s_narrowed" % self.grammar.name

        def process():
            self.engine.set_grammar(self)
        state.add_post_collection_task(process)
        return state



#
# This file is a command-module for Dragonfly.
# (c) Copyright 2008 by Christo Butcher
# Licensed under the LGPL, see <http://www.gnu.org/licenses/>
#

"""
Command-module loader for WSR
=============================

This script can be used to look Dragonfly command-modules 
for use with Window Speech Recognition.  It scans the 
directory it's in and loads any ``*.py`` it finds.

"""


import time
import os.path
import logging
import pythoncom

from dragonfly.engines.backend_sapi5.engine import Sapi5InProcEngine
from dragonfly.loader import CommandModuleDirectory


#---------------------------------------------------------------------------
# Set up basic logging.

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("compound.parse").setLevel(logging.INFO)


#---------------------------------------------------------------------------
# Main event driving loop.

def main():
    logging.basicConfig(level=logging.INFO)

    try:
        path = os.path.dirname(__file__)
    except NameError:
        # The "__file__" name is not always available, for example
        #  when this module is run from PythonWin.  In this case we
        #  simply use the current working directory.
        path = os.getcwd()
        __file__ = os.path.join(path, "dfly-loader-wsr.py")

    engine = Sapi5InProcEngine()
    engine.connect()

    directory = CommandModuleDirectory(path, excludes=[__file__])
    directory.load()

    engine.speak('beginning loop!')
    while 1:
        pythoncom.PumpWaitingMessages()
        time.sleep(.1)

if __name__ == "__main__":
    main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#


import unittest
import time
import logging
from dragonfly.engines import get_engine


#===========================================================================

class CapturingHandler(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)
        self.records = []

    def emit(self, record):
        self.records.append(record)


#===========================================================================

class TestTimer(unittest.TestCase):

    def setUp(self):
        self.log_capture = CapturingHandler()
        logging.getLogger("engine.timer").addHandler(self.log_capture)
        self.engine = get_engine()

    def test_timer_callback_exception(self):
        """ Test handling of exceptions during timer callback. """

        callback_called = [0]
        def callback():
            callback_called[0] += 1
            raise Exception("Exception from timer callback")

        interval = 0.01
        timer = self.engine.create_timer(callback, interval)
        timer.manager.main_callback()
        time.sleep(0.02)
        timer.manager.main_callback()

        self.assertEqual(callback_called, [1])
        self.assertEqual(len(self.log_capture.records), 1)
        log_message = self.log_capture.records[0].msg
        self.assertTrue("Exception from timer callback" in log_message)

#===========================================================================

if __name__ == "__main__":
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
Context factory implementations
============================================================================

"""
from six import string_types

import dragonfly
from .loader    import ContextFactoryBase
from .loader    import ContextFactoryError as Error


#===========================================================================
# Context factory classes.

class GlobalContextFactory(ContextFactoryBase):

    name = "global"

    def build(self, call_info):
        if call_info.arguments:
            raise Error("Global context doesn't need arguments;"
                        " received: %r" % (call_info.arguments,))
        return dragonfly.Context()


#---------------------------------------------------------------------------

class ExecutableContextFactory(ContextFactoryBase):

    name = "executable"

    def build(self, call_info):
        executable = _get_single_anon_string(call_info.arguments,
                                             "Executable context")
        return dragonfly.AppContext(executable=executable)


#---------------------------------------------------------------------------

class TitleContextFactory(ContextFactoryBase):

    name = "title"

    def build(self, call_info):
        title = _get_single_anon_string(call_info.arguments,
                                        "Title context")
        return dragonfly.AppContext(title=title)


#---------------------------------------------------------------------------
# Utility functions.

def _get_single_anon_string(arguments, description):
    if len(arguments) != 1:
        raise Error("%s expects 1 argument,"
                    " received %d." % (description, len(arguments)))
    argument = arguments[0]
    if argument.name is not None:
        raise Error("%s expects 1 argument"
                    " without a name." % description)
    if not isinstance(argument.value, string_types):
        raise Error("%s expects 1 argument"
                    " that is a literal string." % description)
    return argument.value



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it 
#   under the terms of the GNU Lesser General Public License as published 
#   by the Free Software Foundation, either version 3 of the License, or 
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but 
#   WITHOUT ANY WARRANTY; without even the implied warranty of 
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public 
#   License along with Dragonfly.  If not, see 
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements a Win32 dialog base class.
"""


#---------------------------------------------------------------------------

import sys
import ctypes
import struct
import winxpgui as win32gui
import win32api
import win32con
import os
from ctypes.wintypes import POINT


#---------------------------------------------------------------------------

class MINMAXINFO(ctypes.Structure):
    _fields_ = [
                ("ptReserved",      POINT),
                ("ptMaxSize",       POINT),
                ("ptMaxPosition",   POINT),
                ("ptMinTrackSize",  POINT),
                ("ptMaxTrackSize",  POINT),
               ]


#---------------------------------------------------------------------------

class DialogBase(object):

    _dialog_registered_windowclass = False
    _dialog_atom = None

    @classmethod
    def _dialog_register_windowclass(cls, instance, name):
        if not cls._dialog_registered_windowclass:
            wc = win32gui.WNDCLASS()
            wc.SetDialogProc()
            wc.hInstance = instance
            wc.lpszClassName = name
            wc.style = win32con.CS_VREDRAW | win32con.CS_HREDRAW
            wc.hCursor = win32gui.LoadCursor( 0, win32con.IDC_ARROW )
            wc.hbrBackground = win32con.COLOR_WINDOW + 1
            wc.lpfnWndProc = {}

            # C code: wc.cbWndExtra = DLGWINDOWEXTRA + sizeof(HBRUSH) + (sizeof(COLORREF));
            wc.cbWndExtra = win32con.DLGWINDOWEXTRA + struct.calcsize("Pi")

            cls._dialog_atom = win32gui.RegisterClass(wc)
            cls._dialog_registered_windowclass = True

        return cls._dialog_atom


    #-----------------------------------------------------------------------

    def __init__(self, classname, title, initial_size, initial_position,
                 min_size, max_size, style, font, modal=False):

        self._dialog_classname = classname
        self._dialog_initial_size = initial_size
        self._dialog_initial_position = initial_position
        self._dialog_min_size = min_size
        self._dialog_max_size = max_size
        self._dialog_title = title
        self._dialog_font = font
        self._dialog_style = style
        self._dialog_modal = modal

        win32gui.InitCommonControls()
        self._hinst = win32gui.dllhandle
        self._dialog_controls = []

        self._dialog_register_windowclass(self._hinst, classname)

    def _dialog_create(self, template, message_map):
        if self._dialog_modal:  function = win32gui.DialogBoxIndirect
        else:                   function = win32gui.CreateDialogIndirect
        function(self._hinst, template, 0, message_map)

    def _dialog_build_template(self):
        # Initialize dialog template with header.
        dialog_template = [self._dialog_build_header()]

        # Create controls and append control entries to the dialog template.
        self._dialog_build_controls()
        w, h = self._dialog_initial_size
        for control in self._dialog_controls:
            dialog_template.append(control.template_entry(w, h))

        print("\n   ".join(str(value) for value in dialog_template))
        return dialog_template

    def _calculate_center(self, size):
        w, h = size
        desktop = win32gui.GetDesktopWindow()
        al, at, ar, ab = win32gui.GetWindowRect(desktop)
        cx = (ar-al)/2; cy = (ab-at)/2
        cx, cy = win32gui.ClientToScreen(desktop, (cx, cy))
        l = cx - w/2; t = cy - h/2
        return l, t


    #-----------------------------------------------------------------------
    # Property access to window info.

    hwnd = property(lambda self: self._hwnd)


    #-----------------------------------------------------------------------
    # Methods that control the creation and content of the window.

    def _dialog_build_header(self):
        """
            Method which creates window controls.

            This method should be overridden by derived window classes
            to create the desired contents.

            http://docs.activestate.com/activepython/2.5/pywin32/Dialog_Header_Tuple.html

        """

        return [
                self._dialog_title,
                tuple(self._dialog_initial_position)
                 + tuple(self._dialog_initial_size),
                self._dialog_style,
                None,
                self._dialog_font,
                None,
                self._dialog_classname,
               ]

    def _dialog_build_controls(self):
        """
            Method which creates window controls.

            This method should be overridden by derived window classes
            to create the desired contents.

        """

        pass

    def add_control(self, control):
        """
            Register the given *control* as a child of this window.

            This method is generally called by the constructor of
            Control instances.  It is therefore usually not necessary for
            users to call it directly.

        """

        self._dialog_controls.append(control)

    def _dialog_build_message_map(self):
        # Collect all controls which expect callbacks.
        map = {}
        for control in self._dialog_controls:
            for message, callback in list(control.message_callbacks.items()):
                map.setdefault(message, {})[control.id] = callback

        # Create dispatchers for each type of message.
        for message, control_callbacks in list(map.items()):
            def dispatcher(hwnd, msg, wparam, lparam):
                id = win32api.LOWORD(wparam)
                if id in control_callbacks:
                    control_callbacks[id](hwnd, msg, wparam, lparam)
            map[message] = dispatcher

        # Add the top-level callbacks handled by the window itself.
        map.update({
                    win32con.WM_SIZE:           self.on_size,
                    win32con.WM_INITDIALOG:     self.on_init_dialog,
                    win32con.WM_GETMINMAXINFO:  self.on_getminmaxinfo,
                    win32con.WM_CLOSE:          self.on_close,
                    win32con.WM_DESTROY:        self.on_destroy,
                  })
        return map


    #-----------------------------------------------------------------------
    # Message handler methods.

    def _get_edge_sizes(self):
        wl, wt, wr, wb = win32gui.GetWindowRect(self._hwnd)
        cl, ct, cr, cb = win32gui.GetClientRect(self._hwnd)
        il, it = win32gui.ClientToScreen(self._hwnd, (cl, ct))
        ir, ib = win32gui.ClientToScreen(self._hwnd, (cr, cb))
        el = il - wl
        et = it - wt
        er = wr - ir
        eb = wb - ib
        return el, et, er, eb

    def tracer (function):
        if hasattr(function, 'im_func'): name = function.__func__.__name__
        else: name = function.__name__
        def decorated (*a, **k):
            print("call", name, "arguments:", a, k)
            return function(*a, **k)
        return decorated

    @tracer
    def on_init_dialog(self, hwnd, msg, wparam, lparam):
        self._hwnd = hwnd

        l, t = self._dialog_initial_position
        el, et, er, eb = self._get_edge_sizes()
        w = self._dialog_initial_size[0] + el + er
        h = self._dialog_initial_size[1] + et + eb
        win32gui.MoveWindow(self._hwnd, l, t, w, h, 0)

        l, t, r, b = win32gui.GetWindowRect(self._hwnd)
        w = r - l - el - er
        h = b - t - et - eb
        self._do_size(w, h, 1)

        self._post_init()

    def _post_init(self):
        """
            Method for performing any processing required after the
            main window has been initialized.

            This method should be overridden by derived window classes
            if they require anything to be done after the main window
            has been created and positioned.  This method is called
            after a WM_INITDIALOG message was received.

        """

        pass

    def on_size(self, hwnd, msg, wparam, lparam):
        width  = win32api.LOWORD(lparam)
        height = win32api.HIWORD(lparam)
        self._do_size(width, height)
        return 1

    def _do_size(self, width, height, repaint=1):
        for control in self._dialog_controls:
            l, t, w, h = control.calculate_size(width, height)
            win32gui.MoveWindow(control.handle, l, t, w, h, repaint)

    def on_getminmaxinfo(self, hwnd, msg, wparam, lparam):
        info = ctypes.cast(lparam, ctypes.POINTER(MINMAXINFO)).contents
        el, et, er, eb = self._get_edge_sizes()
        if self._dialog_max_size:
            width  = self._dialog_max_size[0] + el + er
            height = self._dialog_max_size[1] + et + eb
            point  = POINT(width, height)
            info.ptMaxSize       = point
            info.ptMaxTrackSize  = point
        if self._dialog_min_size:
            width  = self._dialog_min_size[0] + el + er
            height = self._dialog_min_size[1] + et + eb
            point  = POINT(width, height)
            info.ptMinTrackSize  = point

    def close(self):
        if self._dialog_modal:  win32gui.EndDialog(self._hwnd, 0)
        else:                   win32gui.DestroyWindow(self._hwnd)

    @tracer
    def on_close(self, hwnd, msg, wparam, lparam):
            self.close()

    @tracer
    def on_destroy(self, hwnd, msg, wparam, lparam):
        if not self._dialog_modal:
            win32gui.PostQuitMessage(0)


#---------------------------------------------------------------------------

class BasicDialog(DialogBase):

    def __init__(self, title, size, min_size=None, max_size=None, modal=False):
        classname  = "dragonfly.BasicDialog"
        position   = self._calculate_center(size)
        font       = (8, "MS Sans Serif")
        style      = ( win32con.WS_THICKFRAME
                     | win32con.WS_POPUP
                     | win32con.WS_VISIBLE
                     | win32con.WS_CAPTION
                     | win32con.WS_SYSMENU
                     | win32con.DS_SETFONT
                     | win32con.WS_MINIMIZEBOX)

        DialogBase.__init__(
                            self,
                            classname=classname,
                            title=title,
                            initial_size=size,
                            initial_position=position,
                            min_size=min_size,
                            max_size=max_size,
                            style=style,
                            font=font,
                            modal=modal,
                           )

        template = self._dialog_build_template()
        message_map = self._dialog_build_message_map()
        self._dialog_create(template, message_map)





import time
from six import string_types

import win32file, win32event
import win32con

import logging


#===========================================================================
# Internal container class representing a monitored directory.

class _Dir(object):

    __slots__= ("path", "handle", "overlapped")

    def __init__(self, path):
        if not isinstance(path, string_types):
            raise TypeError("Path argument must be a string/unicode object;"
                            " instead received %r" % (path,))
        self.path = path
        self.handle = win32file.CreateFile(
             self.path,
             0x0001,  # FILE_LIST_DIRECTORY
             win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE,
             None,
             win32con.OPEN_EXISTING,
             win32con.FILE_FLAG_BACKUP_SEMANTICS | win32con.FILE_FLAG_OVERLAPPED,
             None,
            )
        self.overlapped = win32file.OVERLAPPED()
        self.overlapped.hEvent = win32event.CreateEvent(None, True, 0, None)

    def is_modified(self, buffer):
        print("path", self.path)
        result = win32file.ReadDirectoryChangesW(
             self.handle,
             buffer,
             True,
                win32con.FILE_NOTIFY_CHANGE_FILE_NAME 
              | win32con.FILE_NOTIFY_CHANGE_DIR_NAME
              | win32con.FILE_NOTIFY_CHANGE_ATTRIBUTES
              | win32con.FILE_NOTIFY_CHANGE_SIZE
              | win32con.FILE_NOTIFY_CHANGE_LAST_WRITE
              | win32con.FILE_NOTIFY_CHANGE_SECURITY,
             self.overlapped,
            )
        print("result  1", result)
        result = win32event.WaitForMultipleObjects([self.overlapped.hEvent], True, 0)
        print("result  2", result)
        if result != win32event.WAIT_TIMEOUT:
#            result = win32file.GetOverlappedResult(self.handle, self.overlapped, False)
            return True
        else:
            return False


#===========================================================================
# Directory monitor class.

class DirectoryMonitor(object):

    _log = logging.getLogger("dirmon")

    def __init__(self):
        self._buffer = win32file.AllocateReadBuffer(8192)
        self._directories = []

    def add_directories(self, *directories):
        """
            Add one or more directories to be monitored.

            Parameter:
              - directories (sequence of *str* or *unicode*) --
                the new directories to be monitored.

        """
        for path in directories:
            if path not in self.directories:
                try:
                    self._directories.append(_Dir(path))
                except Exception as e:
                    self._log.exception("Failed to monitor directory %r: %s"
                                        % (path, e))

    @property
    def directories(self):
        """ A sequence of paths to the monitored directories. """
        return tuple(d.path for d in self._directories)

    def is_modified(self):
        """ Return True if any of the monitored directories has changed. """
        for directory in self._directories:
            if directory.is_modified(self._buffer):
                return True



# encoding: utf-8
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#
"""
Text action
============================================================================

This section describes the :class:`Text` action object. This type of
action is used for typing text into the foreground application.

It differs from the :class:`Key` action in that :class:`Text` is used for
typing literal text, while :class:`dragonfly.actions.action_key.Key`
emulates pressing keys on the keyboard.  An example of this is that the
arrow-keys are not part of a text and so cannot be typed using the
:class:`Text` action, but can be sent by the
:class:`dragonfly.actions.action_key.Key` action.


Windows Unicode Keyboard Support
............................................................................

The :class:`Text` action can be used to type arbitrary Unicode characters
using the `relevant Windows API <https://docs.microsoft.com/en-us/windows/desktop/api/winuser/ns-winuser-tagkeybdinput#remarks>`__.
This is disabled by default because it ignores the up/down status of
modifier keys (e.g. ctrl).

It can be enabled by changing the ``unicode_keyboard`` setting in
`~/.dragonfly2-speech/settings.cfg` to ``True``::

    unicode_keyboard = True


If you need to simulate typing arbitrary Unicode characters *and* have
*individual* :class:`Text` actions respect modifier keys normally for normal
characters, set the configuration as above and use the ``use_hardware``
parameter for :class:`Text` as follows:

.. code:: python

   Text(u"σμ") + Key("ctrl:down") + Text("]", use_hardware=True) + Key("ctrl:up")


Some applications require hardware emulation versus Unicode keyboard
emulation. If you use such applications, add their executable names to the
``hardware_apps`` list in the configuration file mentioned above to make
dragonfly always use hardware emulation for them.


Text class reference
............................................................................

"""


from six import text_type

from ..engines import get_engine
from ..windows.clipboard import Clipboard
from ..windows.window import Window
from .action_base import ActionError, DynStrActionBase
from .action_key import Key
from .keyboard import Keyboard
from .typeables import typeables

# ---------------------------------------------------------------------------

UNICODE_KEYBOARD = False
HARDWARE_APPS = [
            "tvnviewer.exe", "vncviewer.exe", "mstsc.exe", "virtualbox.exe"
        ]


def load_configuration():
    """Locate and load configuration."""
    import io
    import os
    try:
        import configparser
    except ImportError:
        import ConfigParser as configparser

    global UNICODE_KEYBOARD
    global HARDWARE_APPS

    home = os.path.expanduser("~")
    config_folder = os.path.join(home, ".dragonfly2-speech")
    config_file = "settings.cfg"
    config_path = os.path.join(config_folder, config_file)

    if not os.path.exists(config_folder):
        os.mkdir(config_folder)
    if not os.path.exists(config_path):
        with io.open(config_path, "w") as f:
            # Write the default values to the config file.
            f.write(u'[Text]\n')
            f.write(u'hardware_apps = %s\n' % "|".join(HARDWARE_APPS))
            f.write(u'unicode_keyboard = %s\n' % UNICODE_KEYBOARD)

    parser = configparser.ConfigParser()
    parser.read(config_path)
    if parser.has_option("Text", "hardware_apps"):
        HARDWARE_APPS = parser.get("Text", "hardware_apps").lower().split("|")
    if parser.has_option("Text", "unicode_keyboard"):
        UNICODE_KEYBOARD = parser.getboolean("Text", "unicode_keyboard")


load_configuration()


def require_hardware_emulation():
    """Return `True` if the current context requires hardware emulation."""
    from os.path import basename
    foreground_executable = basename(Window.get_foreground()
                                     .executable.lower())
    return (not UNICODE_KEYBOARD) or (foreground_executable in HARDWARE_APPS)


class Text(DynStrActionBase):
    """
    `Action` that sends keyboard events to type text.

    Arguments:
     - *spec* (*str*) -- the text to type
     - *static* (boolean) --
       if *True*, do not dynamically interpret *spec*
       when executing this action
     - *pause* (*float*) --
       the time to pause between each keystroke, given
       in seconds
     - *autofmt* (boolean) --
       if *True*, attempt to format the text with correct
       spacing and capitalization.  This is done by first mimicking
       a word recognition and then analyzing its spacing and
       capitalization and applying the same formatting to the text.
     - *use_hardware* (boolean) --
       if *True*, send keyboard events using hardware emulation instead of
       as Unicode text. This will respect the up/down status of modifier
       keys.
    """

    _pause_default = 0.02
    _keyboard = Keyboard()
    _specials = {
                 "\n": typeables["enter"],
                 "\t": typeables["tab"],
                }

    def __init__(self, spec=None, static=False, pause=_pause_default,
                 autofmt=False, use_hardware=False):
        self._pause = pause
        self._autofmt = autofmt
        self._use_hardware = use_hardware
        DynStrActionBase.__init__(self, spec=spec, static=static)

    def _parse_spec(self, spec):
        """Convert the given *spec* to keyboard events."""
        from struct import unpack
        events = []
        if self._use_hardware or require_hardware_emulation():
            for character in spec:
                if character in self._specials:
                    typeable = self._specials[character]
                    events.extend(typeable.events(self._pause))
                else:
                    try:
                        typeable = Keyboard.get_typeable(character)
                        events.extend(typeable.events(self._pause))
                    except ValueError:
                        raise ActionError("Keyboard interface cannot type this"
                                          " character: %r (in %r)"
                                          % (character, spec))
        else:
            for character in text_type(spec):
                if character in self._specials:
                    typeable = self._specials[character]
                    events.extend(typeable.events(self._pause))
                else:
                    byte_stream = character.encode("utf-16-le")
                    for short in unpack("<" + str(len(byte_stream) // 2) + "H",
                                        byte_stream):
                        try:
                            typeable = Keyboard.get_typeable(short,
                                                             is_text=True)
                            events.extend(typeable.events(self._pause * 0.5))
                        except ValueError:
                            raise ActionError("Keyboard interface cannot type "
                                              "this character: %r (in %r)" %
                                              (character, spec))
        return events

    def _execute_events(self, events):
        """
        Send keyboard events.

        If instance was initialized with *autofmt* True,
        then this method will mimic a word recognition
        and analyze its formatting so as to autoformat
        the text's spacing and capitalization before
        sending it as keyboard events.
        """
        if self._autofmt:
            # Mimic a word, select and copy it to retrieve capitalization.
            get_engine().mimic("test")
            Key("cs-left, c-c/5").execute()
            word = Clipboard.get_text()

            # Inspect formatting of the mimicked word.
            index = word.find("test")
            if index == -1:
                index = word.find("Test")
                capitalize = True
                if index == -1:
                    self._log.error("Failed to autoformat.")
                    return False
            else:
                capitalize = False

            # Capitalize given text if necessary.
            text = self._spec
            if capitalize:
                text = text[0].capitalize() + text[1:]

            # Reconstruct autoformatted output and convert it
            #  to keyboard events.
            prefix = word[:index]
            suffix = word[index + 4:]
            events = self._parse_spec(prefix + text + suffix)

        # Send keyboard events.
        self._keyboard.send_keyboard_events(events)
        return True



# This Python file uses the following encoding: utf-8

"""
CMU Pocket Sphinx engine configuration module.
"""

from sphinxwrapper import DefaultConfig
from pyaudio import paInt16
import os

# Configuration for the Pocket Sphinx decoder.
DECODER_CONFIG = DefaultConfig()

# Silence the decoder output by default.
DECODER_CONFIG.set_string("-logfn", os.devnull)

# Set voice activity detection configuration options for the decoder.
# You may wish to experiment with these if noise in the background triggers speech
# start and/or false recognitions (e.g. of short words) frequently. Descriptions
# for VAD configuration options were retrieved from:
# https://cmusphinx.github.io/doc/sphinxbase/fe_8h_source.html

# Number of silence frames to keep after from speech to silence.
DECODER_CONFIG.set_int("-vad_postspeech", 50)

# Number of speech frames to keep before silence to speech.
DECODER_CONFIG.set_int("-vad_prespeech", 20)

# Number of speech frames to trigger vad from silence to speech.
DECODER_CONFIG.set_int("-vad_startspeech", 10)

# Threshold for decision between noise and silence frames. Log-ratio between signal
# level and noise level.
DECODER_CONFIG.set_float("-vad_threshold", 3.0)

# Configuration for wake/sleep phrases
# Note that the following CMU Sphinx tutorial has some advice on keyword threshold
# values: https://cmusphinx.github.io/wiki/tutoriallm/#keyword-lists
START_ASLEEP = True
WAKE_PHRASE = "wake up"
WAKE_PHRASE_THRESHOLD = 1e-20
SLEEP_PHRASE = "go to sleep"
SLEEP_PHRASE_THRESHOLD = 1e-40

# Configuration for acoustic model training.
# Relative directory paths will use the current working directory, i.e. where the
# loader is.
TRAINING_DATA_DIR = "training/"
START_TRAINING_PHRASE = "start training session"
START_TRAINING_THRESHOLD = 1e-48
END_TRAINING_PHRASE = "end training session"
END_TRAINING_THRESHOLD = 1e-45

# Keyword arguments given to the PyAudio.open method for opening a stream from a
# microphone. PyAudio streams are used by the engine to recognise speech from audio.
# You may wish to change the values to be optimal for the models you are using;
# Sphinx models require quite specific audio input to recognise speech correctly.
# The values below should work for the default US English models.
PYAUDIO_STREAM_KEYWORD_ARGS = {
    "input": True,              # stream is an input stream
    "format": paInt16,          # 16-bit rate
    "channels": 1,              # one channel (mono)
    "rate": 16000,              # 16kHz sample rate
    "frames_per_buffer": 2048,  # frames per buffer set to 2048
}

# User language for the engine to use.
LANGUAGE = "en"

# Timeout in seconds for speaking the next part of a rule involving dictation.
# This does not include the first part of such a rule. If this value is set to 0,
# then there will be no timeout.
#
# If speech matched any normal rule that doesn't need to be spoken in sequence, it
# will be processed after the timeout period if it happens. Take the following rule
# specs for example:
# "hello"
# "hello <dictation>"
# If you say "hello" and don't start speaking the <dictation> part of the second
# rule, then the first "hello" rule will be processed instead.
NEXT_PART_TIMEOUT = 2.0
